from math import pi, sqrt, asin, cos
import numpy
from numpy import nan
from typing import Optional
from .radiation import get_radiation_integrals, ohmi_envelope
from .energy_loss import get_energy_loss, get_timelag_fromU0
from ..lattice import Lattice, Orbit
from ..constants import clight, Cgamma, Cq

__all__ = ['RingParameters', 'radiation_parameters', 'envelope_parameters']


class RingParameters(object):
    """Class for pretty printing the ring properties"""

    _props = {
        'tunes':            '              Frac. tunes: {0}',
        'tunes6':           '  Frac. tunes (6D motion): {0}',
        'fulltunes':        '                    Tunes: {0}',
        'chromaticities':   '           Chromaticities: {0}',
        'alphac':           ' Momentum compact. factor: {0:e}',
        'etac':             '              Slip factor: {0:e}',
        'E0':               '                   Energy: {0:e} eV',
        'U0':               '       Energy loss / turn: {0:e} eV',
        'i1':               ' Radiation integrals - I1: {0} m',
        'i2':               '                       I2: {0} m^-1',
        'i3':               '                       I3: {0} m^-2',
        'i4':               '                       I4: {0} m^-1',
        'i5':               '                       I5: {0} m^-1',
        'emittances':       '          Mode emittances: {0}',
        'J':                'Damping partition numbers: {0}',
        'Tau':              '            Damping times: {0} s',
        'sigma_e':          '            Energy spread: {0:g}',
        'sigma_l':          '             Bunch length: {0:g} m',
        'voltage':          '         Cavities voltage: {0} V',
        'phi_s':            '        Synchrotron phase: {0:g} rd',
        'f_s':              '    Synchrotron frequency: {0:g} Hz'
    }

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        vrs = vars(self).copy()
        vals = [(self._props[k], vrs.pop(k, None)) for k in self._props]
        # Predefined attributes
        lines = [f.format(v) for f, v in vals if v is not None]
        # Other attributes
        lines += ['{0:>25}: {1}'.format(k, getattr(self, k)) for k in vrs]
        return '\n'.join(lines)


# noinspection PyPep8Naming
def radiation_parameters(ring: Lattice, dp: Optional[float] = None,
                         params: Optional[RingParameters] = None,
                         **kwargs):
    r"""Compute ring parameters from the radiation integrals

    Valid for uncoupled lattices with no RF cavity or radiating element.

    Parameters:
        ring:       Lattice description.
        dp:         Momentum deviation.
        params:     :py:class:`.RingParameters` object to be updated.
          Default: create a new one

    Keyword Args:
        dct (float):        Path lengthening. If specified, ``dp`` is
          ignored and the off-momentum is deduced from the path lengthening.
        method (Callable):  Method for linear optics:

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis
        orbit (Orbit):      Avoids looking for the closed orbit if is
          already known ((6,) array)
        XYStep (float):     Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):     Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        params:             :py:class:`.RingParameters` object.

    **params** is a :py:class:`.RingParameters` object with the following
    attributes:

    ==================  ========================================
    **tunes**           (3,) fractional (H, V, Long.) tunes
    **fulltunes**       (3,) full tunes
    **chromaticities**  (2,) H, V Chromaticities
    **alphac**          Momentum compaction factor
    **etac**            Frequency slip factor
    **E0**              Energy [eV]
    **U0**              Energy loss / turn [eV]
    **i1**              Radiation integrals - :math:`I_1 \quad [m]`
    **i2**              :math:`I_2 \quad [m^{-1}]`
    **i3**              :math:`I_3 \quad [m^{-2}]`
    **i4**              :math:`I_4 \quad [m^{-1}]`
    **i5**              :math:`I_5 \quad [m^{-1}]`
    **emittances**      (3,) Mode emittances
    **J**               (3,) Damping partition numbers
    **Tau**             (3,) Damping times [s]
    **sigma_e**         Energy spread
    **sigma_l**         Bunch length [m]
    **voltage**         Total accelerating voltage [V]
    **phi_s**           Synchrotron phase [rad]
    **f_s**             Synchrotron frequency [Hz]
    ==================  ========================================
    """
    rp = RingParameters() if params is None else params
    _, ringdata, twiss = ring.get_optics(refpts=range(len(ring) + 1), dp=dp,
                                         get_chrom=True, **kwargs)
    rp.chromaticities = ringdata.chromaticity * ring.periodicity
    integs = get_radiation_integrals(ring, dp=dp, twiss=twiss)
    rp.i1, rp.i2, rp.i3, rp.i4, rp.i5 = numpy.array(integs) * ring.periodicity
    circumference = ring.circumference
    voltage = ring.rf_voltage
    E0 = ring.energy
    gamma = ring.gamma
    gamma2 = gamma * gamma
    beta = sqrt(1.0 - 1.0/gamma2)
    U0 = Cgamma / 2.0 / pi * E0**4 * rp.i2
    Jx = 1.0 - rp.i4/rp.i2
    Jz = 1.0
    Je = 2.0 + rp.i4/rp.i2
    damping_partition_numbers = numpy.array([Jx, Jz, Je])
    revolution_period = circumference / clight / beta
    ct = 2.0 * E0 / U0 * revolution_period
    rp.E0 = E0
    rp.U0 = U0
    emitx = Cq * gamma2 * rp.i5 / Jx / rp.i2
    alphac = rp.i1 / circumference
    etac = 1.0/gamma2 - alphac
    rp.phi_s = (pi - asin(U0 / voltage)) if U0 <= voltage else nan
    nus = sqrt(abs(etac * ring.harmonic_number *
               voltage * cos(rp.phi_s) / E0 / 2.0 / pi)) / beta
    rp.voltage = voltage
    rp.f_s = nus / revolution_period
    rp.Tau = ct / damping_partition_numbers
    rp.J = damping_partition_numbers
    rp.sigma_e = sqrt(Cq * gamma2 * rp.i3 / Je / rp.i2)
    rp.sigma_l = beta * abs(etac) * circumference / 2.0 / pi / nus * rp.sigma_e
    rp.emittances = numpy.array([emitx, nan, rp.sigma_e*rp.sigma_l])
    ringtunes, _ = numpy.modf(ring.periodicity * ringdata.tune)
    if len(ringtunes) < 3:
        rp.tunes = numpy.concatenate((ringtunes, (nus,)))
    else:
        rp.tunes6 = ringtunes
    rp.fulltunes = ring.periodicity * twiss[-1].mu / 2.0 / pi
    rp.alphac = alphac
    rp.etac = etac
    return rp


# noinspection PyPep8Naming
def envelope_parameters(ring: Lattice,
                        params: Optional[RingParameters] = None,
                        orbit: Orbit = None,
                        keep_lattice: bool = False) -> RingParameters:
    r"""Compute ring parameters from ohmi_envelope

    Parameters:
        ring:           Lattice description.
        params:         :py:class:`.RingParameters` object to be updated.
          Default: create a new one
        orbit:          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        keep_lattice:   Assume no lattice change since the
          previous tracking.

    Returns:
        params:             :py:class:`.RingParameters` object.

    **params** is a :py:class:`.RingParameters` object with the following
    attributes:

    ==================  ========================================
    **tunes6**          (3,) fractional (H, V, Long.) tunes (6D motion)
    **emittances**      (3,) Mode emittances
    **J**               (3,) Damping partition numbers
    **Tau**             (3,) Damping times [s]
    **sigma_e**         Energy spread
    **sigma_l**         Bunch length [m]
    **voltage**         Total accelerating voltage [V]
    **phi_s**           Synchrotron phase [rad]
    **f_s**             Synchrotron frequency [Hz]
    ==================  ========================================
    """
    rp = RingParameters() if params is None else params
    emit0, beamdata, emit = ohmi_envelope(ring, orbit=orbit,
                                          keep_lattice=keep_lattice)
    voltage = ring.rf_voltage
    rp.E0 = ring.energy
    rp.U0 = get_energy_loss(ring)
    rev_freq = ring.revolution_frequency
    rp.Tau = 1.0 / rev_freq / beamdata.damping_rates / ring.periodicity
    alpha = 1.0 / rp.Tau
    rp.J = 4.0 * alpha / numpy.sum(alpha)
    rp.tunes6, _ = numpy.modf(ring.periodicity * beamdata.tunes)
    rp.phi_s = pi - numpy.arcsin(rp.U0 / voltage)
    rp.voltage = voltage
    rp.f_s = rp.tunes6[2] * rev_freq
    rp.emittances = beamdata.mode_emittances
    rp.sigma_e = sqrt(emit0.r66[4, 4])
    rp.sigma_l = sqrt(emit0.r66[5, 5])
    return rp


Lattice.radiation_parameters = radiation_parameters
Lattice.envelope_parameters = envelope_parameters



import matlab
import numpy as np
import pytest
from numpy.testing import assert_allclose as assert_close

from at import find_orbit4, find_sync_orbit, find_orbit6
from at import uint32_refpts, find_m44, find_m66


def _py_data(ml_data):
    """Convert a Matlab vector to a numpy vector"""
    return np.squeeze(np.asarray(ml_data))


def _ml_refs(refpts, nelems):
    """Convert refpoints to Matlab"""
    uintrefs = uint32_refpts(refpts, nelems)
    return matlab.double([float(ref + 1) for ref in uintrefs])


@pytest.mark.parametrize("ct", (-0.0002, 0.0002))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_find_syncorbit(engine, request, lattices, ct):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)

    # Python call
    py_orb5, py_orbit5 = find_sync_orbit(py_lattice, ct, refpts)
    # Matlab call
    ml_orbit5, ml_orb5 = engine.findsyncorbit(
        ml_lattice, ct, _ml_refs(refpts, nelems), nargout=2
    )
    ml_orbit5 = np.rollaxis(np.asarray(ml_orbit5), -1)
    # Comparison
    assert_close(py_orb5, _py_data(ml_orb5), atol=1.0e-12, rtol=0)
    assert_close(py_orbit5[:, :5], ml_orbit5, atol=1.0e-12, rtol=0)


@pytest.mark.parametrize("dp", (-0.01, 0.01))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_find_orbit4(engine, request, lattices, dp):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)

    # Python call
    py_orb4, py_orbit4 = find_orbit4(py_lattice, dp, refpts)
    # Matlab call
    ml_orbit4, ml_orb4 = engine.findorbit4(
        ml_lattice, dp, _ml_refs(refpts, nelems), nargout=2
    )
    ml_orbit4 = np.rollaxis(np.asarray(ml_orbit4), -1)
    # Comparison
    assert_close(py_orb4, _py_data(ml_orb4), atol=1e-15, rtol=0)
    assert_close(py_orbit4[:, :4], ml_orbit4, atol=1.5e-15, rtol=0)


@pytest.mark.parametrize("lattices", ["hmba_rad"])
def test_find_orbit6(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)

    # Python call
    py_orb6, py_orbit6 = find_orbit6(py_lattice, refpts)
    # Matlab call
    ml_orbit6, ml_orb6 = engine.findorbit6(
        ml_lattice, _ml_refs(refpts, nelems), nargout=2
    )
    ml_orbit6 = np.rollaxis(np.asarray(ml_orbit6), -1)
    # Comparison
    assert_close(py_orb6, _py_data(ml_orb6), atol=5e-15, rtol=0)
    assert_close(py_orbit6, ml_orbit6, atol=5e-15, rtol=0)


@pytest.mark.parametrize("dp", (-0.01, 0.0, 0.01))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_find_m44(engine, request, lattices, dp):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)
    morbit = engine.zeros(6, 1)
    porbit = np.zeros((6, 1))

    # Python call
    py_m44, py_mstack = find_m44(py_lattice, dp, refpts, orbit=porbit)
    # Matlab call
    ml_m44, ml_mstack = engine.findm44(
        ml_lattice, dp, _ml_refs(refpts, nelems), "orbit", morbit, nargout=2
    )
    ml_mstack = np.rollaxis(np.asarray(ml_mstack).reshape((4, 4, nrefs)), -1)
    # Comparison
    assert_close(py_m44, np.asarray(ml_m44), atol=1.0e-11, rtol=0)
    assert_close(py_mstack, ml_mstack, atol=1.0e-11, rtol=0)


@pytest.mark.parametrize("lattices", ["hmba_rad"])
def test_find_m66_orb(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)
    morbit = engine.zeros(6, 1)
    porbit = np.zeros((6, 1))

    # Python call
    py_m66, py_mstack = find_m66(py_lattice, refpts, orbit=porbit)
    # Matlab call
    ml_m66, ml_mstack = engine.findm66(
        ml_lattice, _ml_refs(refpts, nelems), "orbit", morbit, nargout=2
    )
    assert_close(py_m66, np.asarray(ml_m66), atol=1.0e-11, rtol=0)

    ml_mstack = np.rollaxis(np.asarray(ml_mstack).reshape((6, 6, nrefs)), -1)
    assert_close(py_mstack, ml_mstack, atol=1.0e-11, rtol=0)


@pytest.mark.parametrize("lattices", ["hmba_rad"])
def test_find_m66(engine, request, lattices):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    nrefs = nelems + 1
    refpts = range(nrefs)

    # Python call
    py_m66, py_mstack = find_m66(py_lattice, refpts)
    # Matlab call
    ml_m66, ml_mstack = engine.findm66(ml_lattice, _ml_refs(refpts, nelems), nargout=2)
    assert_close(py_m66, np.asarray(ml_m66), atol=1e-7, rtol=0)

    ml_mstack = np.rollaxis(np.asarray(ml_mstack).reshape((6, 6, nrefs)), -1)
    assert_close(py_mstack, ml_mstack, atol=1e-7, rtol=0)



"""momentum aperture alternative."""

from __future__ import annotations

import time
from typing import Dict

import numpy

from ..lattice.lattice_object import Lattice

__all__ = ["momaperture_project2start", "projectrefpts"]

# 2024may29 oblanco at ALBA CELLS. First working version
#           Based on the MATLAB implementation by Z.Marti at ALBA
#           See https://github.com/atcollab/at/pull/773


def momaperture_project2start(ring: Lattice, **kwargs: Dict[str, any]) -> numpy.ndarray:
    """
    :py:func:`momap_project2start` calculates the local momemtum aperture.

    It is a binary search of the negative and positive momentum thresholds
    of stability around the closed orbit.

    For a given momentum offset the particles are first tracked from every
    reference point to the end of the ring, and then, all or a set of particles
    with different 6D coordinates are tracked together.  The surviving particles
    continue the boundary search with a new energy step until the boundary is
    found when the step limit or the momentum limit is reached.

    Usage:
      >>> momaperture_project2start(ring)

    Parameters:
      ring: list of elements

    Keyword Arguments:
      refpts: Selects the location of coordinates output.
        See ":ref:`Selecting elements in a lattice <refpts>`"
      nturns: number of turns to be tracked. Default 1000
      dptol: momentum offset resolution. Default 1e-4
      dpuguess: maximum momentum boundary. Default: the rf bucket height.
      troffset: (2, N) offsets to be added to the transverse coordinates
        on the N reference points. Default 1e-5 m
      orbit: (N,6) offsets to be added on the N reference points.
        Default, the closed orbit
      verbose: print in the standard output additional info. Default False
      epsilon6d: float.
        If not passed, all particles are tracked.
        If epsilon6d is given, we track for nturns only particles
        having 6D coordinates different by epsilon6d

    Returns:
      dnp: (N,2) array with negative and positive stable momentum boundaries
        for the N reference points

    ..note::
      * This function could track in parallel. Set use_mp=True.
        Other arguments could be passed, check :py:func:`lattice_track`.
      * This function does a quick search, but, it is succeptible to miss
        islands of instability due to the varying energy step.
    """
    # verboseprint to check flag only once
    verbose = kwargs.pop("verbose", False)
    verboseprint = print if verbose else lambda *a, **k: None

    rps = kwargs.pop("refpts", ring.uint32_refpts(range(len(ring))))
    rps = ring.get_uint32_index(rps)
    nrps = len(rps)
    verboseprint(f"Using {nrps} reference points")

    nturns = kwargs.pop("nturns", 1000)
    verboseprint(f"Track over {nturns} turns")

    dptol = kwargs.pop("dptol", 1e-4)
    verboseprint(f"Momentum resolution {dptol}")

    # set transverse offsets
    if "troffset" in kwargs:
        add_offset = kwargs.pop("troffset")
        verboseprint("Add user offsets")
    else:
        dxy = 1e-5
        add_offset = numpy.tile(dxy, [2, nrps])
        verboseprint(f"Adding default transverse offsets {dxy} per plane")

    # set the minimum distance btw particles that makes them similar
    epsilon6d = kwargs.pop("epsilon6d", 0)

    # first guess
    if "dpuguess" in kwargs:
        eu_ini = kwargs.pop("dpuguess")
        verboseprint(f"Using the users max boundary {eu_ini}")
    else:
        # use radiation parameters to get the rf bucket
        pars = ring.radiation_parameters()
        # get energy bucket (deltabucket) max height.
        # S.Y.Lee, 4th Edition, Eqs 3.37, 3.38, Sect. II.2 Bucket area
        harmonnumber = ring.harmonic_number
        etac = pars.etac
        theenergy = pars.E0
        phis = pars.phi_s
        betar = ring.beta
        thevoltage = ring.get_rf_voltage()
        yfactor = numpy.sqrt(
            numpy.abs(numpy.cos(phis) - 0.5 * (numpy.pi - 2 * phis) * numpy.sin(phis))
        )
        deltabucket = (
            numpy.sqrt(
                2
                * thevoltage
                / (numpy.pi * betar**2 * theenergy * harmonnumber * numpy.abs(etac))
            )
            * yfactor
        )
        verboseprint(f"Bucket height {deltabucket}")
        eu_ini = deltabucket
        verboseprint("Using the bucket height as maximum boundary")
    es_ini = 0
    et_ini = eu_ini / 2

    if "orbit" in kwargs:
        orbit_s = kwargs.pop("orbit")
        verboseprint("Using the users orbit")
    else:
        _, orbit_s = ring.find_orbit(rps)
        verboseprint("Using the closed orbit")

    # start scan of:
    # unstable energy
    # stable energy
    # test energy
    etpos = et_ini * numpy.ones(nrps)
    eupos = eu_ini * numpy.ones(nrps)
    espos = es_ini * numpy.ones(nrps)
    etneg = -1 * et_ini * numpy.ones(nrps)
    euneg = -1 * eu_ini * numpy.ones(nrps)
    esneg = -1 * es_ini * numpy.ones(nrps)
    deltae = 1
    iteration = 0
    while iteration < 100 and (deltae > dptol):  # safety limit on iterations
        iteration = iteration + 1
        t00 = time.time()
        # plost is a mask, True for lost particles
        plost = multirefpts_track_islost(
            ring,
            rps,
            etpos,
            etneg,
            orbit_s,
            add_offset,
            nturns,
            epsilon6d,
            verbose,
            **kwargs,
        )
        # split in positive and negative side of energy offsets
        plostpos = plost[0::2]
        plostneg = plost[1::2]
        # split in stable (es), unstable (eu) and test (et) energy
        espos[~plostpos] = etpos[~plostpos]
        eupos[plostpos] = etpos[plostpos]
        etpos = (espos + eupos) / 2
        esneg[~plostneg] = etneg[~plostneg]
        euneg[plostneg] = etneg[plostneg]
        etneg = (esneg + euneg) / 2
        # define new energy step
        depos = max(abs(espos - eupos))
        deneg = max(abs(esneg - euneg))
        deltae = max(depos, deneg)
        outmsg = (
            f"Iteration {iteration}",
            f" took {format(time.time()-t00):.3} s.",
            f" dp_step={deltae}, dptol={dptol}",
        )
        verboseprint("".join(outmsg))
    return numpy.vstack([etneg, etpos]).T


def projectrefpts(
    ring: Lattice,
    startrefpts: numpy.ndarray,
    particles: numpy.ndarray,
    **kwargs: Dict[str, any],
) -> tuple:
    """
    :py:fun:`projectrefpts` tracks from multiple reference points.

    Usage:
      >>> projectrefpts(ring, startrefpts, particles)

    Parameters:
      ring: list of elements
      startrefpts: reference point to start tracking from.
      particles: (6,N,R,1) array, where N is the number of
        particles per reference point, and R is the number of
        reference points.

    Keyword arguments:
      endrefpt: end reference point. Default: end of last ring element
      use_mp: Default True. See :py:fun:`lattice_track`
      group: Default False. The starting point info is removed.
        All tracked particles are grouped together.
      verbose: prints additional info

    Returns:
      zout:     Tracked particle coordinates at the end ref. point.
                Default (6,N,R,1) array, where N is the number of
                  particles per R references points
                If the flag 'group' is used the output becomes a
                  (6,N*R,1,1) array with all particles.
      lostpart: Bool array, True when the particle is lost.
                Default (N,R).
                If the flag 'group' is used the output becomes a
                  (N*R) array
    """
    lenring = len(ring)
    rps = startrefpts
    nrps = len(rps)
    nparticles = 1
    if len(particles.shape) >= 2:
        nparticles = particles.shape[1]

    # verboseprint to check flag only once
    verbose = kwargs.pop("verbose", False)
    verboseprint = print if verbose else lambda *a, **k: None

    if "endrefpt" in kwargs:
        erps = kwargs["endrefpt"]
        verboseprint(f"Project particles to start of element index {erps}")
    else:
        erps = lenring
        verboseprint("Project to end point")

    groupparts = kwargs.pop("group", False)

    verboseprint(f"nparticles={nparticles} per reference point")
    verboseprint(f"Number of reference points {nrps}")

    # default to parallel
    use_mp = kwargs.pop("use_mp", True)
    if nparticles == 1:
        use_mp = False

    if groupparts:
        zout = numpy.zeros((6, nparticles * nrps, 1, 1))
        lostpart = numpy.ones((nparticles * nrps), dtype=bool)
    else:
        zout = numpy.zeros((6, nparticles, nrps, 1))
        lostpart = numpy.ones((nparticles, nrps), dtype=bool)

    # first, track the remaining portion of the ring
    zin = particles.copy()
    for i in range(nrps):
        ring_downstream = ring.rotate(rps[i])
        zaux = numpy.squeeze(zin[:, :, i, 0])
        verboseprint(
            f"Tracking {nparticles} particles on reference point {i+1} of {nrps}"
        )
        zoaux, _, dout1 = ring_downstream.track(
            zaux, nturns=1, refpts=erps - rps[i], losses=True, use_mp=use_mp
        )
        if groupparts:
            zout[:, nparticles * i : nparticles * (i + 1), 0, 0] = numpy.reshape(
                zoaux, (6, nparticles)
            )
            lostpart[nparticles * i : nparticles * (i + 1)] = dout1["loss_map"][
                "islost"
            ]
        else:
            zout[:, :, i, 0] = numpy.reshape(zoaux, (6, nparticles))
            lostpart[:, i] = dout1["loss_map"]["islost"]
    return zout, lostpart


def multirefpts_track_islost(
    ring: Lattice,
    refpts: numpy.ndarray,
    esetptpos: numpy.ndarray,
    esetptneg: numpy.ndarray,
    orbit: numpy.ndarray,
    initcoord: numpy.ndarray,
    nturns: float,
    epsilon6d: float,
    verbose: bool,
    **kwargs: Dict[str, any],
) -> numpy.ndarray:
    """
    Tell whether the particle launched is lost.

    Usage:
      >>> multirefpts_track_islost(ring, refpts, energysetpt, orbit, initcoord)

    Parameters:
      ring: list of elements

    Keyword Arguments:
      refpts: Selects the locations.
      esetptpos: positive energy set point for tracking.
      esetptneg: negative energy set point for tracking.
      orbit: (6,N) orbit to be added to the N refpts.
      initcoords: (2,N) hor. and ver. transverse offsets in m.
      nturns: number of turns to track
      epsilon6d: maximum value to consider particles as similar
      verbose: print info

    Returns:
      Lostpart: (2*N) bool array, for N reference points.
        True if the particle is lost.
    """
    verboseprint = print if verbose else lambda *a, **k: None

    # track positive and negative side at the same time
    nparticles = 2

    rps = refpts
    nrps = len(rps)
    lostpart = numpy.ones((nparticles * nrps), dtype=bool)
    zin = numpy.zeros((6, nparticles * nrps))
    zout = numpy.zeros((6, nparticles * nrps))
    tinyoffset = epsilon6d

    # project to the end of the ring
    erps = len(ring)

    # first, track the remaining portion of the ring
    zin[:, 0::2] = orbit.T.copy()
    zin[:, 1::2] = orbit.T.copy()
    zin[0, 0::2] = zin[0, 0::2] + initcoord[0, :]
    zin[0, 1::2] = zin[0, 1::2] + initcoord[0, :]
    zin[2, 0::2] = zin[2, 0::2] + initcoord[1, :]
    zin[2, 1::2] = zin[2, 1::2] + initcoord[1, :]
    zin[4, 0::2] = zin[4, 0::2] + esetptpos
    zin[4, 1::2] = zin[4, 1::2] + esetptneg
    for i in range(nrps):
        ring_downstream = ring.rotate(rps[i])
        zaux = zin[:, (nparticles * i) : (nparticles * (i + 1))]
        # verboseprint(
        #    f"Tracking {nparticles} particles on reference point {i+1} of {nrps}"
        # )
        zoaux, _, doutaux = ring_downstream.track(
            zaux, nturns=1, refpts=erps - rps[i], losses=True, **kwargs
        )
        zout[:, (nparticles * i) : (nparticles * (i + 1))] = numpy.reshape(
            zoaux, (6, nparticles)
        )
        lostpart[(nparticles * i) : (nparticles * (i + 1))] = doutaux["loss_map"][
            "islost"
        ]

    cntalive = len(lostpart) - sum(lostpart)
    zinaliveaux = zout[:, ~lostpart]
    zinalive_at_ring_end = numpy.asfortranarray(zinaliveaux.copy())

    # second, use the particles that have survived the ring to the end
    # filter them if necessary, and track them
    shapealiveatend = numpy.shape(zinalive_at_ring_end)
    trackonly_mask = numpy.ones(shapealiveatend[1], dtype=bool)
    similarparticles_index = numpy.array([])
    particles_were_filtered = False
    if epsilon6d != 0 and cntalive > 1:
        particles_were_filtered = True
        # search for non numerically similar particles
        closenessmatrix = numpy.zeros((cntalive, cntalive), dtype=bool)
        for i in range(cntalive):
            closenessmatrix[i, :] = [
                numpy.allclose(
                    zinalive_at_ring_end[:, j],
                    zinalive_at_ring_end[:, i],
                    atol=tinyoffset,
                )
                for j in range(cntalive)
            ]
        _, rowidx = numpy.indices((cntalive, cntalive))
        maxidx = numpy.max(rowidx * closenessmatrix, 1)
        trackonly_mask, _, similarparticles_index = numpy.unique(
            maxidx, return_index=True, return_inverse=True
        )
        outmsg = (
            "Speed up when discarding similar particles, ",
            f"{100*len(trackonly_mask)/cntalive:.3f}%",
        )
        verboseprint("".join(outmsg))
    # track
    # dummy track to later reuse the ring
    zaux2 = numpy.array([initcoord[0, 0], 0, initcoord[1, 0], 0, 1e-6, 0]).T
    ring.track(zaux2, nturns=1)
    # track non-numerically similar particles
    _, _, dout_multiturn = ring.track(
        zinalive_at_ring_end[:, trackonly_mask],
        nturns=nturns,
        keep_lattice=True,
        losses=True,
        **kwargs,
    )
    if particles_were_filtered:
        lostpaux = dout_multiturn["loss_map"]["islost"][similarparticles_index]
    else:
        lostpaux = dout_multiturn["loss_map"]["islost"]
    lostpart[~lostpart] = lostpaux

    return lostpart



"""Generic function to save and load python AT lattices. The format is
determined by the file extension
"""

from __future__ import annotations

__all__ = ["load_lattice", "save_lattice", "register_format"]

import os.path
from collections.abc import Callable
from typing import Optional

from at.lattice import Lattice

_load_extension = {}
_save_extension = {}


def load_lattice(filepath: str, **kwargs) -> Lattice:
    """Load a Lattice object from a file

    The file format is indicated by the filepath extension. The file name is stored in
    the *in_file* Lattice attribute. The selected variable, if relevant, is stored
    in the *use* Lattice attribute.

    Parameters:
        filepath:           Name of the file

    Keyword Args:
        use (str):          Name of the variable containing the desired lattice.
          Default: if there is a single variable, use it, otherwise select ``"RING"``
        name (str):         Name of the lattice.
          Default: taken from the file, or ``""``
        energy (float):     Energy of the lattice
          (default: taken from the file)
        periodicity (int):  Number of periods
          (default: taken from the file, or 1)
        *:                  All other keywords will be set as :py:class:`.Lattice`
          attributes

    Returns:
        lattice (Lattice):          New :py:class:`.Lattice` object

    Check the format-specific function for specific keyword arguments:

    .. Admonition:: Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        load_func = _load_extension[ext.lower()]
    except KeyError:
        print("File load failed: unknow extension {}.".format(ext))
    else:
        return load_func(filepath, **kwargs)


def save_lattice(ring: Lattice, filepath: str, **kwargs) -> None:
    """Save a Lattice object

    The file format is indicated by the filepath extension.

    Parameters:
        ring:               Lattice description
        filepath:           Name of the file

    Check the format-specific function for specific keyword arguments:

    .. Admonition:: Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        save_func = _save_extension[ext.lower()]
    except KeyError:
        print("File save failed: unknow extension {}.".format(ext))
    else:
        return save_func(ring, filepath, **kwargs)


def register_format(
    extension: str,
    load_func: Optional[Callable[..., Lattice]] = None,
    save_func: Optional[Callable[..., None]] = None,
    descr: str = "",
):
    """Register format-specific processing functions

    Parameters:
        extension:      File extension string.
        load_func:      load function.
        save_func:      save function.
        descr:          File type description.
    """
    if load_func is not None:
        _load_extension[extension] = load_func
        load_lattice.__doc__ += f"\n        {extension:<10}\n            {descr}\n"
    if save_func is not None:
        _save_extension[extension] = save_func
        save_lattice.__doc__ += f"\n        {extension:<10}\n            {descr}\n"


Lattice.load = staticmethod(load_lattice)
Lattice.save = save_lattice



import numpy
from ..lattice import Lattice
# noinspection PyProtectedMember
from ..lattice.elements import Element, Collective, _array
from ..constants import clight, qe
from .wake_object import Wake, WakeComponent


# noinspection PyPep8Naming
class WakeElement(Collective, Element):
    """Class to generate an AT wake element using the passmethod WakeFieldPass
    """
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES
    default_pass = {False: 'IdentityPass', True: 'WakeFieldPass'}
    _conversions = dict(Element._conversions, _nslice=int, _nturns=int,
                        _nelem=int, _wakeFact=float,
                        NormFact=lambda v: _array(v, (3,)),
                        ZCuts=lambda v: _array(v),
                        _wakeDX=lambda v: _array(v),
                        _wakeDY=lambda v: _array(v),
                        _wakeQX=lambda v: _array(v),
                        _wakeQY=lambda v: _array(v),
                        _wakeZ=lambda v: _array(v),
                        _wakeT=lambda v: _array(v))

    def __init__(self, family_name: str, ring: Lattice, wake: Wake, **kwargs):
        """
        Parameters:
            family name:    Element name
            ring:           Lattice in which the element will be inserted
            wake:           :py:class:`.Wake` object

        Keyword Arguments:
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (Tuple[float,...]):    Normalization for the 3 planes,
              to account for beta function at the observation point for
              example. Default: (1,1,1)
"""
        kwargs.setdefault('PassMethod', self.default_pass[True])
        zcuts = kwargs.pop('ZCuts', None)
        self._wakefact = - ring.circumference/(clight *
                                               ring.energy*ring.beta**3)
        self._nslice = kwargs.pop('Nslice', 101)
        self._nturns = kwargs.pop('Nturns', 1)
        self._turnhistory = None    # Defined here to avoid warning
        self.clear_history(ring=ring)
        self.NormFact = kwargs.pop('NormFact', numpy.ones(3, order='F'))
        self._build(wake)
        if zcuts is not None:
            self.ZCuts = zcuts
        super(WakeElement, self).__init__(family_name, **kwargs)

    def _build(self, wake):
        self._wakeT = wake.srange
        self._nelem = len(self._wakeT)
        if wake.Z is not None:
            self._wakeZ = wake.Z
        if wake.DX is not None:
            self._wakeDX = wake.DX
        if wake.DY is not None:
            self._wakeDY = wake.DY
        if wake.QX is not None:
            self._wakeQX = wake.QX
        if wake.QY is not None:
            self._wakeQY = wake.QY

    def rebuild_wake(self, wake):
        self._build(wake)

    def clear_history(self, ring=None):
        if ring is not None:
            self._nbunch = ring.nbunch
        tl = self._nturns*self._nslice*self._nbunch
        self._turnhistory = numpy.zeros((tl, 4), order='F')

    def set_normfactxy(self, ring):
        l0, _, _ = ring.get_optics()
        self.NormFact[0] = 1/l0['beta'][0]
        self.NormFact[1] = 1/l0['beta'][1]

    @property
    def WakeT(self):
        return self._wakeT

    @property
    def WakeZ(self):
        """Longitudinal component"""
        return getattr(self, '_wakeZ', None)

    @property
    def WakeDX(self):
        """Dipole X component"""
        return getattr(self, '_wakeDX', None)

    @property
    def WakeDY(self):
        """Dipole Y component"""
        return getattr(self, '_wakeDY', None)

    @property
    def WakeQX(self):
        """Quadrupole X component"""
        return getattr(self, '_wakeQX', None)

    @property
    def WakeQY(self):
        """Quadrupole Y component"""
        return getattr(self, '_wakeQY', None)

    @property
    def Nslice(self):
        """Number of slices per bunch"""
        return self._nslice

    @Nslice.setter
    def Nslice(self, nslice):
        self._nslice = nslice
        self.clear_history()

    @property
    def Nturns(self):
        """Number of turn for the wake field"""
        return self._nturns

    @Nturns.setter
    def Nturns(self, nturns):
        self._nturns = nturns
        self.clear_history()

    @property
    def TurnHistory(self):
        """Turn histroy of the slices center of mass"""
        return self._turnhistory

    def __repr__(self):
        """Simplified __repr__ to avoid errors due to arguments
        not defined as attributes
        """
        attrs = dict((k, v) for (k, v) in self.items()
                     if not k.startswith('_'))
        return '{0}({1})'.format(self.__class__.__name__, attrs)


class ResonatorElement(WakeElement):
    """Class to generate a resonator, inherits from WakeElement
       additonal argument are frequency, qfactor, rshunt
    """
    def __init__(self, family_name: str, ring: Lattice, srange,
                 wakecomp: WakeComponent,
                 frequency: float, qfactor: float, rshunt: float,
                 yokoya_factor=1, **kwargs):
        r"""
        Parameters:
            family name:    Element name
            ring:           Lattice in which the element will be inserted
            srange:         Vector of s position where to sample the wake
            wakecomp:       Wake component
            frequency:      Resonator frequency [Hz]
            qfactor:        Q factor
            rshunt:         Shunt impedance, [:math:`\Omega`] for longitudinal,
              [:math:`\Omega/m`] for transverse
            yokoya_factor:  Yokoya factor

        Keyword Arguments:
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (Tuple[float,...]):    Normalization for the 3 planes,
              to account for beta function at the observation point for
              example. Default: (1,1,1)
"""
        self._resfrequency = frequency
        self._qfactor = qfactor
        self._rshunt = rshunt
        self._yokoya = yokoya_factor
        self._wakecomponent = wakecomp
        self._beta = ring.beta
        wake = Wake.resonator(srange, wakecomp, frequency, qfactor, rshunt,
                              ring.beta, yokoya_factor=yokoya_factor)
        super(ResonatorElement, self).__init__(family_name, ring, wake,
                                               **kwargs)

    def rebuild_wake(self):
        wake = Wake.resonator(self.WakeT, self._wakecomponent,
                              self._resfrequency, self._qfactor,
                              self._rshunt, self._beta, self._yokoya)
        self._build(wake)

    @property
    def ResFrequency(self):
        """Resonator frequency [Hz]"""
        return self._resfrequency

    @ResFrequency.setter
    def ResFrequency(self, frequency):
        self._resfrequency = frequency
        self.rebuild_wake()

    @property
    def Qfactor(self):
        """Resonator Q factor"""
        return self._qfactor

    @Qfactor.setter
    def Qfactor(self, qfactor):
        self._qfactor = qfactor
        self.rebuild_wake()

    @property
    def Rshunt(self):
        r"""Resonator shunt impedance, [:math:`\Omega`] for longitudinal,
        [:math:`\Omega/m`] for transverse"""
        return self._rshunt

    @Rshunt.setter
    def Rshunt(self, rshunt):
        self._rshunt = rshunt
        self.rebuild_wake()

    @property
    def Yokoya(self):
        """Resonator Yokoya factor"""
        return self._yokoya

    @Yokoya.setter
    def Yokoya(self, yokoya):
        self._yokoya = yokoya
        self.rebuild_wake()


class LongResonatorElement(ResonatorElement):
    """Class to generate a longitudinal resonator, inherits from WakeElement
       additional argument are frequency, qfactor, rshunt
    """
    def __init__(self, family_name: str, ring: Lattice, srange,
                 frequency: float, qfactor: float, rshunt: float,
                 **kwargs):
        r"""
        Parameters:
            family name:    Element name
            ring:           Lattice in which the element will be inserted
            srange:         Vector of s position where to sample the wake
            frequency:      Resonator frequency [Hz]
            qfactor:        Q factor
            rshunt:         Shunt impedance, [:math:`\Omega`] for longitudinal,
              [:math:`\Omega/m`] for transverse

        Keyword Arguments:
            yokoya_factor:  Yokoya factor
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (Tuple[float,...]):    Normalization for the 3 planes,
              to account for beta function at the observation point for
              example. Default: (1,1,1)
"""
        super(LongResonatorElement, self).__init__(family_name, ring, srange,
                                                   WakeComponent.Z, frequency,
                                                   qfactor, rshunt, **kwargs)


class ResWallElement(WakeElement):
    """Class to generate a resistive wall element, inherits from WakeElement
       additional argument are yokoya_factor, length, pipe radius, conductivity
    """
    def __init__(self, family_name: str, ring: Lattice, srange,
                 wakecomp: WakeComponent, rwlength: float,
                 rvac: float, conduc: float, yokoya_factor=1, **kwargs):
        """
        Parameters:
            family name:    Element name
            ring:           Lattice in which the element will be inserted
            srange:         Vector of s position where to sample the wake
            wakecomp:       Wake component
            rwlength:
            rvac:
            conduc:
            yokoya_factor:  Yokoya factor

        Keyword Arguments:
            Nslice (int):       Number of slices per bunch. Default: 101
            Nturns (int):       Number of turn for the wake field. Default: 1
            ZCuts:              Limits for fixed slicing, default is adaptive
            NormFact (Tuple[float,...]):    Normalization for the 3 planes,
              to account for beta function at the observation point for
              example. Default: (1,1,1)
        """
        self._wakecomponent = wakecomp
        self._rwlength = rwlength
        self._rvac = rvac
        self._conductivity = conduc
        self._yokoya = yokoya_factor
        self._beta = ring.beta
        wake = Wake.resistive_wall(srange, wakecomp, rwlength, rvac, conduc,
                                   ring.beta, yokoya_factor=yokoya_factor)
        super(ResWallElement, self).__init__(family_name, ring, wake, **kwargs)

    def rebuild_wake(self):
        wake = Wake.resistive_wall(self.WakeT, self._wakecomponent,
                                   self._rwlength, self._rvac,
                                   self._conductivity, self._beta,
                                   yokoya_factor=self._yokoya)
        self._build(wake)

    @property
    def RWLength(self):
        """Length of the resistive wall"""
        return self._rwlength

    @RWLength.setter
    def RWLength(self, length):
        self._rwlength = length
        self.rebuild_wake()

    @property
    def Conductivity(self):
        """Conductivity of the beam pipe [S/m]"""
        return self._conductivity

    @Conductivity.setter
    def Conductivity(self, conduct):
        self._conductivity = conduct
        self.rebuild_wake()

    @property
    def Rvac(self):
        """Radius of the beam pipe [m]"""
        return self._rvac

    @Rvac.setter
    def Rvac(self, rvac):
        self._rvac = rvac
        self.rebuild_wake()

    @property
    def Yokoya(self):
        """Yokoya factor for the reistive wall"""
        return self._yokoya

    @Yokoya.setter
    def Yokoya(self, yokoya):
        self._yokoya = yokoya
        self.rebuild_wake()



"""AT plotting functions"""
from ..lattice import Lattice, Refpts
from ..tracking import internal_lpass
from ..physics import get_optics, Orbit
from .generic import baseplot

# --------- Example 1 --------

# The specific data generating functions do not depend on any graphics-related
# function. So they could be located in any module, for instance
# in lattice/linear.py where the get_optics function is.


# data generating function
def pldata_beta_disp(ring: Lattice, refpts: Refpts, **kwargs):
    """Generates data for plotting beta functions and dispersion"""

    # compute linear optics at the required locations
    data0, _, data = get_optics(ring, refpts=refpts, **kwargs)

    # Extract the plot data
    s_pos = data['s_pos']
    betax = data['beta'][:, 0]
    betaz = data['beta'][:, 1]
    dispersion = data['dispersion'][:, 0]

    # Left axis definition
    left = (r'$\beta$ [m]', s_pos, [betax, betaz],
            [r'$\beta_x$', r'$\beta_z$'])
    # Right axis definition
    right = ('dispersion [m]', s_pos, [dispersion], ['dispersion'])
    return 'Optical functions', left, right


# Convenience function to make the call simpler
def plot_beta(ring: Lattice, **kwargs):
    """Plot beta functions and dispersion

    Parameters:
        ring:               Lattice description.

    Keyword Args:
        dp (float):         Momentum deviation.
        dct (float):        Path lengthening. If specified, ``dp`` is
          ignored and the off-momentum is deduced from the path lengthening.
        orbit (Orbit):      Avoids looking for the closed orbit if is
          already known ((6,) array)
        method (Callable):  Method for linear optics (see
          :py:func:`.get_optics`):

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt4`: no longitudinal motion, Sagan/Rubin
          4D-analysis of coupled motion,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):     Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):     Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:           Initial conditions for transfer line optics. Record
          array as output by :py:func:`.linopt2`, :py:func:`.linopt6`, or
          dictionary.
        s_range:            Lattice range of interest, default: unchanged,
          initially set to the full cell.
        axes (tuple[Axes, Optional[Axes]): :py:class:`~matplotlib.axes.Axes`
          for plotting as (primary_axes, secondary_axes).
          Default: create new axes
        slices (int):       Number of slices. Default: 400
        legend (bool):      Show a legend on the plot
        labels (Refpts):    display the name of selected elements.
          Default: :py:obj:`None`
        block (bool):       If :py:obj:`True`, block until the figure is closed.
          Default: :py:obj:`False`
        dipole (dict):      Dictionary of properties overloading the default
          properties of dipole representation. See :py:func:`.plot_synopt`
          for details
        quadrupole (dict):  Same definition as for dipole
        sextupole (dict):   Same definition as for dipole
        multipole (dict):   Same definition as for dipole
        monitor (dict):     Same definition as for dipole
   """
    return baseplot(ring, pldata_beta_disp, **kwargs)

# --------- Example 2 --------


# data generating function
def pldata_linear(ring: Lattice, refpts: Refpts, *keys, **kwargs):
    """data extraction function for plotting results of get_optics"""

    class Lind(object):
        """helper class for lindata extraction"""
        lab2 = "xz"
        lab6 = ("x", "x'", "z", "z'", "l", "\\delta")
        id6 = "123456"
        params = dict(
            beta=(r'$\beta$ [m]', r'$\beta_{0}$', lab2),
            closed_orbit=('position [m]', r'${0}$', lab6),
            dispersion=('dispersion [m]', r'$\eta_{0}$', lab6),
            alpha=(r'$\alpha$', r'$\alpha_{0}$', lab2),
            mu=(r'Phase advance', r'$\mu_{0}$', lab2),
            gamma=('Gamma', 'Gamma', None),
            M=('Transfert', r'$T_{{{0},{1}}}$', id6)
        )

        @classmethod
        def extract(cls, lindata, key, *idx):
            def it(v):
                try:
                    return iter(v)
                except TypeError:
                    return iter([v])

            axis_title, fmt, convert = cls.params[key]
            indices = list(zip(*(it(i) for i in idx)))
            print(indices)
            datay = (lindata[key][(slice(None),) + ic] for ic in indices)
            labels = (fmt.format(*(convert[i] for i in ic)) for ic in indices)
            return axis_title, lindata['s_pos'], datay, labels

    title = kwargs.pop('title', 'Linear optics')
    data0, _, data = get_optics(ring, refpts=refpts, get_chrom=True, **kwargs)
    return (title,) + tuple(Lind.extract(data, *key) for key in keys)


# Convenience function to make the call simpler
def plot_linear(ring: Lattice, *keys, **kwargs):
    """Plot linear optical functions returned by :py:func:`.get_optics`

    Parameters:
        ring:           Lattice description.

    Keyword Args:
        left:           Left axis description as a tuple:
          (key[, indices[, indices]])

          **key**:        'beta', 'closed_orbit',... See :py:func:`.get_optics`

          **indices**:    integer, sequence of integers, or slice

          The number if sequences of indices is data[key].ndim-1
          The number of indices is the number of curves to plot.
          All sequences must have the same length.

        Examples:

          :code:`('beta', [0, 1])`:         beta_x, beta_z

          :code:`('dispersion', 0)`:        eta_x

          :code:`('closed_orbit', [1, 3])`:    x', z'

          :code:`('m44', 2, 2)`:            T33

          :code:`('m44', [0, 0], [0, 1])`:  T11, T12

          :code:`('m44', 2, slice(4))`:     T31, T32, T33, T34 as a single block

          :code:`('m44', [2,2,2,2], [0,1,2,3])`:    T31, T32, T33, T34

        right:              Right axis (optional). See ``left``
        title (str):        Plot title, defaults to "Linear optics"
        dp (float):         Momentum deviation.
        dct (float):        Path lengthening. If specified, ``dp`` is
          ignored and the off-momentum is deduced from the path lengthening.
        orbit (Orbit):      Avoids looking for the closed orbit if is
          already known ((6,) array)
        method (Callable):  Method for linear optics (see
          :py:func:`.get_optics`):

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt4`: no longitudinal motion, Sagan/Rubin
          4D-analysis of coupled motion,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):     Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):     Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:           Initial conditions for transfer line optics. Record
          array as output by :py:func:`~.linear.linopt2`,
          :py:func:`~.linear.linopt6`, or dictionary.
        s_range:            Lattice range of interest, default: unchanged,
          initially set to the full cell.
        axes (tuple[Axes, Optional[Axes]): :py:class:`~matplotlib.axes.Axes`
          for plotting as (primary_axes, secondary_axes).
          Default: create new axes
        slices (int):       Number of slices. Default: 400
        legend (bool):      Show a legend on the plot
        labels (Refpts):    display the name of selected elements.
          Default: :py:obj:`None`
        block (bool):       If :py:obj:`True`, block until the figure is closed.
          Default: :py:obj:`False`
        dipole (dict):      Dictionary of properties overloading the default
          properties of dipole representation. See :py:func:`.plot_synopt`
          for details
        quadrupole (dict):  Same definition as for dipole
        sextupole (dict):   Same definition as for dipole
        multipole (dict):   Same definition as for dipole
        monitor (dict):     Same definition as for dipole
    """
    return baseplot(ring, pldata_linear, *keys, **kwargs)


# --------- Example 3 --------

# Here the data generating function is embedded in the convenience function
def plot_trajectory(ring: Lattice, r_in, nturns: int = 1, **kwargs):
    """Plot a particle's trajectory

    Parameters:
        ring:           Lattice object
        r_in:           (6,n) array: input coordinates of n particles
        nturns:         Number of turns

    Keyword Args:
        keep_lattice (bool):   Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        s_range:            Lattice range of interest, default: unchanged,
          initially set to the full cell.
        axes (tuple[Axes, Optional[Axes]): :py:class:`~matplotlib.axes.Axes`
          for plotting as (primary_axes, secondary_axes).
          Default: create new axes
        slices (int):       Number of slices. Default: 400
        legend (bool):      Show a legend on the plot
        block (bool):       If :py:obj:`True`, block until the figure is closed.
          Default: :py:obj:`False`
        dipole (dict):      Dictionary of properties overloading the default
          properties of dipole representation. See :py:func:`.plot_synopt`
          for details
        quadrupole (dict):  Same definition as for dipole
        sextupole (dict):   Same definition as for dipole
        multipole (dict):   Same definition as for dipole
        monitor (dict):     Same definition as for dipole
    """
    # noinspection PyShadowingNames
    def pldata_trajectory(ring, refpts, r_in, nturns=1, **kwargs):
        r_out = internal_lpass(ring, r_in, refpts=refpts, nturns=nturns,
                               **kwargs)
        s_pos = ring.get_s_pos(refpts)
        particles = range(r_out.shape[1])
        xx = [r_out[0, i, :, :] for i in particles]
        zz = [r_out[2, i, :, :] for i in particles]
        xlabels = [r'$x_{0}$'.format(i) for i in particles]
        zlabels = [r'$z_{0}$'.format(i) for i in particles]
        left = ('position [m]', s_pos, xx+zz, xlabels+zlabels)
        return 'Trajectory', left

    return baseplot(ring, pldata_trajectory, r_in, nturns=nturns, **kwargs)


Lattice.plot_beta = plot_beta
Lattice.plot_trajectory = plot_trajectory



"""
Collection of functions to fit various global parameters
such as the tune and the chromaticity
"""
from typing import Optional
import numpy
from ..lattice import Lattice, Refpts
from ..lattice import get_value_refpts, set_value_refpts
from ..lattice import AtWarning, AtError
from ..physics import get_tune, get_chrom


__all__ = ['fit_tune', 'fit_chrom']


def _set_magnets(ring, refpts, attname, delta, index=None,
                 increment=True, regex=False, scaling=False):
    if scaling:
        val0 = get_value_refpts(ring, refpts, attname, index=index,
                                regex=regex)
        set_value_refpts(ring, refpts, attname, val0*(1+delta/numpy.mean(val0)),
                         index=index, regex=regex)                        
    else:
        set_value_refpts(ring, refpts, attname, delta, index=index,
                         increment=True, regex=regex)
                 


def _get_tune(ring: Lattice, dp: float, **kwargs):
    get_integer = kwargs.pop('fit_integer', False)
    return get_tune(ring, dp=dp, get_integer=get_integer)[0:2]


def _get_chrom(ring: Lattice, dp: float, **kwargs):
    return get_chrom(ring, dp=dp)[0:2]


def _fit_tune_chrom(ring: Lattice, index: int, func,
                    refpts1: Refpts, refpts2: Refpts, newval,
                    tol: Optional[float] = 1.0e-12,
                    dp: Optional[float] = 0, niter: Optional[int] = 3,
                    delta: Optional[float] = None,
                    regex=False, scaling=False, **kwargs):

    def _get_resp(ring: Lattice, index: int, func, refpts, attname,
                  delta, dp, regex=False, scaling=False, **kwargs):
        _set_magnets(ring, refpts, attname, delta, index=index,
                     increment=True, regex=regex, scaling=scaling)
        datap = func(ring, dp, **kwargs)
        _set_magnets(ring, refpts, attname, -2*delta, index=index,
                     increment=True, regex=regex, scaling=scaling)
        datan = func(ring, dp, **kwargs)
        _set_magnets(ring, refpts, attname, delta, index=index,
                     increment=True, regex=regex, scaling=scaling)
        data = numpy.subtract(datap, datan)/(2*delta)
        return data

    def _fit(ring, index, func, refpts1, refpts2, newval, J,
             dp: Optional[float] = 0, regex=False, **kwargs):
        val = func(ring, dp, **kwargs)
        dk = numpy.linalg.solve(J, numpy.subtract(newval, val))
        _set_magnets(ring, refpts1, 'PolynomB', dk[0], index=index,
                     increment=True, regex=regex, scaling=scaling)
        _set_magnets(ring, refpts2, 'PolynomB', dk[1], index=index,
                     increment=True, regex=regex, scaling=scaling)
        val = func(ring, dp, **kwargs)
        sumsq = numpy.sum(numpy.square(numpy.subtract(val, newval)))
        return sumsq

    if delta is None:
        delta = 1.e-6 * 10 ** index
        
    dq1 = _get_resp(ring, index, func, refpts1, 'PolynomB',
                    delta, dp, regex=regex, **kwargs)
    dq2 = _get_resp(ring, index, func, refpts2, 'PolynomB',
                    delta, dp, regex=regex, **kwargs)
    J = [[dq1[0], dq2[0]], [dq1[1], dq2[1]]]

    n = 0
    sumsq = tol+1
    print('Initial value', func(ring, dp, **kwargs))
    while sumsq > tol and n < niter:
        sumsq = _fit(ring, index, func, refpts1, refpts2, newval,
                     J, dp=dp, regex=regex, **kwargs)
        print('iter#', n, 'Res.', sumsq)
        n += 1
    print('Final value', func(ring, dp, **kwargs), '\n')
    return


def fit_tune(ring: Lattice, refpts1: Refpts, refpts2: Refpts, newval,
             tol: float = 1.0e-12,
             dp: Optional[float] = 0, niter: int = 3, regex=False, 
             KStep: Optional[float] = None,
             **kwargs) -> None:
    """Fits the tunes using 2 families

    Args:
        ring:       Lattice description
        refpts1:    Selection of the 1st family
        refpts2:    Selection of the 2nd family
        newval:     New tunes, in case an non-zero integer part
                    is provided, fit_integer is set to True

    Keyword arguments:
        tol:        Tolerance for the matching; Default: 1.0e-12
        dp:         Momentum deviation. Default: 0
        niter:      Maximum number of iterations. Default 3
        fit_integer: bool (default=False), use integer tune
        regex:      Using regular expressions for refpt string matching;
                    Default: False
        KStep: gradient variation applied to magnets. Default 1e-5
        scaling: Scales the families instead of incrementing.
                 Default False

    Typical usage:
    at.fit_tune(ring, refpts1, refpts2, [0.1,0.25])
    """
    print('\nFitting Tune...')
    if numpy.any(numpy.floor(newval) != 0.0):
        kwargs['fit_integer'] = True
    _fit_tune_chrom(ring, 1, _get_tune, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter, regex=regex, delta=KStep, **kwargs)


def fit_chrom(ring: Lattice, refpts1: Refpts, refpts2: Refpts, newval,
              tol: Optional[float] = 1.0e-12,
              dp: Optional[float] = 0, niter: Optional[int] = 3, regex=False,
              HStep: Optional[float] = None,
              **kwargs) -> None:
    """Fit the chromaticities using 2 families

    Args:
        ring:       Lattice description
        refpts1:    Selection of the 1st family
        refpts2:    Selection of the 2nd family
        newval:     New tunes

    Keyword arguments:
        tol:        Tolerance for the matching; Default: 1.0e-12
        dp:         Momentum deviation. Default: 0
        niter:      Maximum number of iterations. Default 3
        regex:      Using regular expressions for refpt string matching;
                    Default: False
        HStep: gradient variation applied to magnets. Default 1e-4
        scaling: Scales the families instead of incrementing them.
                 Default False

    Typical usage:
    at.fit_chrom(ring, refpts1, refpts2, [10,5])
    """
    print('\nFitting Chromaticity...')
    _fit_tune_chrom(ring, 2, _get_chrom, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter, regex=regex, delta=HStep, **kwargs)



from at import elements
from at import lattice_track
from at import patpass, internal_plpass
import numpy
import pytest
# import sys


# @pytest.mark.skipif(not sys.platform.startswith("win"),
#                    reason="May hang on linux and MacOS")
@pytest.mark.parametrize('func', (lattice_track, patpass, internal_plpass))
def test_patpass_multiple_particles_and_turns(func):
    nturns = 10
    nparticles = 10
    rin = numpy.zeros((6, nparticles))
    d = elements.Drift('drift', 1.0)
    lattice = [d]
    rin[1, 0] = 1e-6
    rin[3, 0] = -2e-6
    if func == lattice_track:
        rout, *_ = func(lattice, rin, nturns, use_mp=True)
    else:
        rout = func(lattice, rin, nturns)
    # results from Matlab
    assert rout.shape == (6, nparticles, 1, nturns)
    rout_expected = numpy.array([1e-6, 1e-6, -2e-6, -2e-6, 0,
                                 2.5e-12]).reshape(6, 1)
    zeros = numpy.zeros((6,))
    # The fourth index is for nturns; the first element is after one turn.
    numpy.testing.assert_equal(rout[:, 0, :, 0], rout_expected)
    for i in range(nturns):
        # The first particle is not all zeros.
        with pytest.raises(AssertionError):
            numpy.testing.assert_equal(rout[:, 0, 0, i], zeros)
        # All other particles are all zeros.
        for j in range(1, nparticles):
            numpy.testing.assert_equal(rout[:, j, 0, i], zeros)



"""
Utility functions for tracking simulations
"""
from __future__ import annotations
import numpy
import functools
from collections.abc import Sequence, Iterable
from typing import Optional
from ..lattice import Lattice, Element
from ..lattice import BeamMoments, Collective, QuantumDiffusion
from ..lattice import SimpleQuantDiff, VariableMultipole
from ..lattice import elements, refpts_iterator, set_value_refpts
from ..lattice import DConstant, checktype, checkattr, get_bool_index


__all__ = ['fortran_align', 'get_bunches', 'format_results',
           'get_bunches_std_mean', 'unfold_beam', 'has_collective',
           'initialize_lpass', 'disable_varelem', 'variable_refs']


DIMENSION_ERROR = 'Input to lattice_pass() must be a 6xN array.'
_COLLECTIVE_ELEMS = (BeamMoments, Collective)
_VAR_ELEMS = (QuantumDiffusion, SimpleQuantDiff, VariableMultipole)
_DISABLE_ELEMS = _COLLECTIVE_ELEMS + _VAR_ELEMS


def _set_beam_monitors(ring: Sequence[Element], nbunch: int, nturns: int):
    """Function to initialize the beam monitors"""
    monitors = list(refpts_iterator(ring, elements.BeamMoments))
    monitors += list(refpts_iterator(ring, elements.SliceMoments))
    for m in monitors:
        m.set_buffers(nturns, nbunch)  
    return len(monitors) == 0


def variable_refs(ring):
    idref = get_bool_index(ring, checkattr('PassMethod', 'IdentityPass'))
    varref = get_bool_index(ring, checktype(_DISABLE_ELEMS))
    varref = varref & ~idref
    return varref


def has_collective(ring) -> bool:
    """True if any element involves collective effects"""
    refs = get_bool_index(ring, checktype(_COLLECTIVE_ELEMS))
    return sum(refs) > 0


def disable_varelem(ring):
    """Function to disable collective effects elements"""
    refs = variable_refs(ring)
    if sum(refs) > 0:
        ring = set_value_refpts(ring, refs, 'PassMethod',
                                'IdentityPass', copy=True)
    return ring  


def _get_bunch_config(lattice, unfoldbeam):
    """Function to get the bunch configuration"""
    nbunch = getattr(lattice, 'nbunch', 1)
    bunch_currents = getattr(lattice, 'bunch_currents', numpy.zeros(1))
    if unfoldbeam:
        bunch_spos = getattr(lattice, 'bunch_spos', numpy.zeros(1))
    else:
        bunch_spos = numpy.zeros(len(bunch_currents))
    return nbunch, bunch_spos, bunch_currents


def initialize_lpass(lattice: Iterable[Element], nturns: int,
                     kwargs) -> list[Element]:
    """Function to initialize keyword arguments for lattice tracking"""
    if not isinstance(lattice, list):
        lattice = list(lattice)
    unfoldbeam = kwargs.pop('unfold_beam', True)
    nbunch, bspos, bcurrents = _get_bunch_config(lattice, unfoldbeam)
    kwargs.update(bunch_currents=bcurrents, bunch_spos=bspos)
    no_bm = _set_beam_monitors(lattice, nbunch, nturns)
    kwargs['keep_lattice'] = kwargs.get('keep_lattice', False) and no_bm
    pool_size = kwargs.pop('pool_size', None)
    start_method = kwargs.pop('start_method', None)
    if kwargs.get('use_mp', False):
        kwargs['pool_size'] = pool_size
        kwargs['start_method'] = start_method
    return lattice


def fortran_align(func):
    # noinspection PyShadowingNames
    """decorator to ensure that *r_in* is Fortran-aligned

    :py:func:`fortran_align` ensures that the 2nd argument (usually *r_in*) of
    the decorated function is Fortran-aligned before calling the function

    Example:

        >>> @fortran_align
        ... def element_pass(element: Element, r_in, **kwargs):
        ... ...

        Ensure that *r_in* is Fortran-aligned
    """
    @functools.wraps(func)
    def wrapper(lattice, r_in, *args, **kwargs):
        assert r_in.shape[0] == 6 and r_in.ndim in (1, 2), DIMENSION_ERROR
        if r_in.flags.f_contiguous:
            return func(lattice, r_in, *args, **kwargs)
        else:
            r_fin = numpy.asfortranarray(r_in)
            r_out = func(lattice, r_fin, *args, **kwargs)
            r_in[:] = r_fin[:]
            return r_out
    return wrapper


def format_results(results, r_in, losses):
    """Function to format the ouput of parallelized tracking"""
    lin, lout = zip(*results)
    # Update r_in with values at the end of tracking
    numpy.concatenate(lin, out=r_in, axis=1)
    if losses:
        lout, ldic = zip(*lout)
        keys = ldic[0].keys()
        dicout = dict(((k, numpy.hstack([li[k] for li in ldic]))
                      for k in keys))
        return numpy.concatenate(lout, axis=1), dicout
    else:
        return numpy.concatenate(lout, axis=1)


def get_bunches(r_in: numpy.ndarray, nbunch: int,
                selected_bunches: Optional[numpy.ndarray] = None):
    """Function to get the bunches particles 6D coordinates

    Parameters:
      r_in: 6 x n_particles Fortran-ordered numpy array.
      nbunch: integer, total number of bunches
      selected_bunches: integer or array of integers, index
      of the selected bunches

    Returns:
      List of ndarray containing the 6 x n particles coordinates
      of the selected bunches
    """
    if selected_bunches is None:
        selected_bunches = numpy.arange(nbunch)
    else:
        selected_bunches = numpy.atleast_1d(selected_bunches)
    bunches = [r_in.T[ib::nbunch].T for ib in selected_bunches]
    return bunches


def get_bunches_std_mean(r_in: numpy.ndarray, nbunch: int,
                         selected_bunches: Optional[numpy.ndarray] = None):
    """Function to get the bunches standard deviation and center
    of mass

    Parameters:
      r_in: 6 x n_particles Fortran-ordered numpy array.
      nbunch: integer, total number of bunches
      selected_bunches: integer or array of integers, index
      of the selected bunches

    Returns:
      Lists of ndarray containing the 6D standard deviation
      and center of mass (std, mean)
    """
    bunches = get_bunches(r_in, nbunch, selected_bunches)
    std = [numpy.nanstd(b, axis=1) for b in bunches]
    mean = [numpy.nanmean(b, axis=1) for b in bunches]
    return std, mean


def unfold_beam(ring: Lattice, beam: numpy.ndarray,
                **kwargs) -> numpy.ndarray:
    """Function to unfold the beam based on the ring fill pattern.
    The input particle distribution has to be in on bucket 0.
    Particle are distributed in bunches using ``i%ring.nbunch``
    where i is the particle index.
    For each bunches the absolute ``ct`` is computed using the 6D
    closed orbit search, this closed orbit is added to the input
    particles.

    Parameters:
        ring: Lattice description
        beam: array with shape(6, nparticles)

    Keyword Arguments:
        convergence (float):    Convergence criterion for 6D orbit
          Default: :py:data:`DConstant.OrbConvergence <.DConstant>`
        max_iterations (int):   Maximum number of iterations for
          6D orbit.
          Default: :py:data:`DConstant.OrbMaxIter <.DConstant>`
    Return:
        beam (numpy.ndarray): unfolded beam
    """
    conv = kwargs.pop('convergence', DConstant.OrbConvergence)
    maxiter = kwargs.pop('max_iterations', DConstant.OrbMaxIter)
    o60, _ = ring.find_orbit(max_iterations=maxiter, convergence=conv)
    unfolded_beam = numpy.zeros(beam.shape)
    for i, spos in enumerate(ring.bunch_spos[::-1]):
        guess = o60.copy()
        guess[5] -= spos
        o6, _ = ring.find_orbit(guess=guess, max_iterations=maxiter,
                                convergence=conv)
        unfolded_beam[:, i::ring.nbunch] = (beam[:, i::ring.nbunch].T + o6).T
    return unfolded_beam



"""
Module to define common elements used in AT.

Each element has a default PassMethod attribute for which it should have the
appropriate attributes.  If a different PassMethod is set, it is the caller's
responsibility to ensure that the appropriate attributes are present.
"""

from __future__ import annotations

import abc
import re
from abc import ABC
from collections.abc import Generator, Iterable
from copy import copy, deepcopy
from typing import Any, Optional

import numpy

# noinspection PyProtectedMember
from .variables import _nop


def _array(value, shape=(-1,), dtype=numpy.float64):
    # Ensure proper ordering(F) and alignment(A) for "C" access in integrators
    return numpy.require(value, dtype=dtype, requirements=['F', 'A']).reshape(
        shape, order='F')


def _array66(value):
    return _array(value, shape=(6, 6))


def _float(value) -> float:
    return float(value)


def _int(value, vmin: Optional[int] = None, vmax: Optional[int] = None) -> int:
    intv = int(value)
    if vmin is not None and intv < vmin:
        raise ValueError(f"Value must be greater of equal to {vmin}")
    if vmax is not None and intv > vmax:
        raise ValueError(f"Value must be smaller of equal to {vmax}")
    return intv


class LongtMotion(ABC):
    """Abstract Base class for all Element classes whose instances may modify
    the particle momentum

    Allows to identify elements potentially inducing longitudinal motion.

    Subclasses of :py:class:`LongtMotion` must provide two methods for
    enabling longitudinal motion:

    * ``_get_longt_motion(self)`` must return the activation state,
    * ``set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs)``
      must enable or disable longitudinal motion.
    """

    @abc.abstractmethod
    def _get_longt_motion(self):
        return False

    # noinspection PyShadowingNames
    @abc.abstractmethod
    def set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs):
        """Enable/Disable longitudinal motion

        Parameters:
            enable:     :py:obj:`True`: for enabling, :py:obj:`False` for
              disabling
            new_pass:   New PassMethod:

              * :py:obj:`None`: makes no change,
              * ``'auto'``: Uses the default conversion,
              * Anything else is used as the new PassMethod.
            copy:       If True, returns a modified copy of the element,
              otherwise modifies the element in-place
        """
        # noinspection PyUnresolvedReferences
        if new_pass is None or new_pass == self.PassMethod:
            return self if copy else None
        if copy:
            newelem = deepcopy(self)
            newelem.PassMethod = new_pass
            return newelem
        # noinspection PyAttributeOutsideInit
        self.PassMethod = new_pass


# noinspection PyUnresolvedReferences
class _DictLongtMotion(LongtMotion):
    # noinspection PyShadowingNames
    """Mixin class for elements implementing a 'default_pass' class attribute

    :py:class:`DictLongtMotion` provides:

    * a :py:meth:`set_longt_motion` method setting the PassMethod according
      to the ``default_pass`` dictionary.
    * a :py:obj:`.longt_motion` property set to :py:obj:`True` when the
      PassMethod is ``default_pass[True]``

    The class must have a ``default_pass`` class attribute, a dictionary
    such that:

    * ``default_pass[False]`` is the PassMethod when radiation is turned
      OFF,
    * ``default_pass[True]`` is the default PassMethod when radiation is
      turned ON.

    The :py:class:`DictLongtMotion` class must be set as the first base class.

    Example:

        >>> class QuantumDiffusion(_DictLongtMotion, Element):
        ...
        ...     default_pass = {False: 'IdentityPass', True: 'QuantDiffPass'}

        Defines a class such that :py:meth:`set_longt_motion` will select
        ``'IdentityPass'`` or ``'IdentityPass'``.
    """

    def _get_longt_motion(self):
        return self.PassMethod != self.default_pass[False]

    # noinspection PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, **kwargs):
        if new_pass == 'auto':
            new_pass = self.default_pass[enable]
        return super().set_longt_motion(enable, new_pass=new_pass, **kwargs)


# noinspection PyUnresolvedReferences
class _Radiative(LongtMotion):
    # noinspection PyShadowingNames
    r"""Mixin class for radiating elements

    :py:class:`_Radiative` implements the mechanism for converting the pass
    methods of radiating elements. It provides:

    * a :py:meth:`set_longt_motion` method setting the PassMethod
      according to the following rule:

      * ``enable == True``: replace "\*Pass" by "\*RadPass"
      * ``enable == False``: replace "\*RadPass" by "\*Pass"
    * a :py:obj:`.longt_motion` property set to true when the PassMethod
      ends with "RadPass"

    The :py:class:`_Radiative` class must be set as the first base class.

    Example:
        >>> class Multipole(_Radiative, LongElement, ThinMultipole):

        Defines a class where :py:meth:`set_longt_motion` will convert the
        PassMethod according to the \*Pass or \*RadPass suffix.
    """

    def _get_longt_motion(self):
        return self.PassMethod.endswith(('RadPass', 'QuantPass'))

    def _autopass(self, enable):
        if enable:
            root = self.PassMethod.replace('QuantPass', 'Pass').\
                replace('RadPass', 'Pass')
            return ''.join((root[:-4], 'RadPass'))
        elif self.longt_motion:
            root = self.PassMethod.replace('QuantPass', 'Pass').\
                replace('RadPass', 'Pass')
            return root
        else:
            return None

    # noinspection PyTypeChecker,PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs):
        if new_pass == 'auto':
            new_pass = self._autopass(enable)
        if new_pass is None or new_pass == self.PassMethod:
            return self if copy else None
        if enable:

            def setpass(el):
                el.PassMethod = new_pass
                el.Energy = kwargs['energy']

        else:

            def setpass(el):
                el.PassMethod = new_pass
                try:
                    del el.Energy
                except AttributeError:
                    pass

        if copy:
            newelem = deepcopy(self)
            setpass(newelem)
            return newelem
        setpass(self)


class Radiative(_Radiative):
    # noinspection PyShadowingNames
    r"""Mixin class for default radiating elements (:py:class:`.Dipole`,
    :py:class:`.Quadrupole`, :py:class:`.Wiggler`)

    :py:class:`Radiative` is a base class for the subset of radiative elements
    considered as the ones to be turned on by default: :py:class:`.Dipole`,
    :py:class:`.Quadrupole` and :py:class:`.Wiggler`, excluding the higher
    order multipoles.

    :py:class:`Radiative` inherits from :py:class:`_Radiative` and does not
    add any new functionality. Its purpose is to identify the default set of
    radiating elements.

    Example:
        >>> class Dipole(Radiative, Multipole):

        Defines a class belonging to the default radiating elements. It
        converts the PassMethod according to the "\*Pass" or "\*RadPass"
        suffix.
    """
    pass


class Collective(_DictLongtMotion):
    """Mixin class for elements representing collective effects

    Derived classes will automatically set the
    :py:attr:`~Element.is_collective` property when the element is active.

    The class must have a ``default_pass`` class attribute, a dictionary such
    that:

    * ``default_pass[False]`` is the PassMethod when collective effects
      are turned OFF,
    * ``default_pass[True]`` is the default PassMethod when collective effects
      are turned ON.

    The :py:class:`Collective` class must be set as the first base class.

    Example:
        >>> class WakeElement(Collective, Element):
        ...
        ...     default_pass = {False: 'IdentityPass', True: 'WakeFieldPass'}

        Defines a class where the :py:attr:`~Element.is_collective` property is
        handled
    """

    def _get_collective(self):
        # noinspection PyUnresolvedReferences
        return self.PassMethod != self.default_pass[False]

    @abc.abstractmethod
    def clear_history(self):
        pass


class Element(object):
    """Base class for AT elements"""

    _BUILD_ATTRIBUTES = ['FamName']
    _conversions = dict(FamName=str, PassMethod=str, Length=_float,
                        R1=_array66, R2=_array66,
                        T1=lambda v: _array(v, (6,)),
                        T2=lambda v: _array(v, (6,)),
                        RApertures=lambda v: _array(v, (4,)),
                        EApertures=lambda v: _array(v, (2,)),
                        KickAngle=lambda v: _array(v, (2,)),
                        PolynomB=_array, PolynomA=_array,
                        BendingAngle=_float,
                        MaxOrder=_int, NumIntSteps=lambda v: _int(v, vmin=0),
                        Energy=_float,
                        )

    _entrance_fields = ['T1', 'R1']
    _exit_fields = ['T2', 'R2']
    _no_swap = _entrance_fields + _exit_fields

    def __init__(self, family_name: str, **kwargs):
        """
        Parameters:
            family_name:    Name of the element

        All keywords will be set as attributes of the element
        """

        self.FamName = family_name
        self.Length = kwargs.pop('Length', 0.0)
        self.PassMethod = kwargs.pop('PassMethod', 'IdentityPass')
        self.update(kwargs)

    def __setattr__(self, key, value):
        try:
            value = self._conversions.get(key, _nop)(value)
        except Exception as exc:
            exc.args = ('In element {0}, parameter {1}: {2}'.format(
                self.FamName, key, exc),)
            raise
        else:
            super(Element, self).__setattr__(key, value)

    def __str__(self):
        return "\n".join(
            [self.__class__.__name__ + ":"]
            + [f"{k:>14}: {v!s}" for k, v in self.items()]
        )

    def __repr__(self):
        clsname, args, kwargs = self.definition
        keywords = [f"{arg!r}" for arg in args]
        keywords += [f"{k}={v!r}" for k, v in kwargs.items()]
        args = re.sub(r"\n\s*", " ", ", ".join(keywords))
        return f"{clsname}({args})"

    def equals(self, other) -> bool:
        """Whether an element is equivalent to another.

        This implementation was found to be too slow for the generic
        __eq__ method when comparing lattices.
        """
        return repr(self) == repr(other)

    def divide(self, frac) -> list[Element]:
        """split the element in len(frac) pieces whose length
        is frac[i]*self.Length

        Parameters:
            frac:           length of each slice expressed as a fraction of the
              initial length. ``sum(frac)`` may differ from 1.

        Returns:
            elem_list:  a list of elements equivalent to the original.

        Example:

            >>> Drift('dr', 0.5).divide([0.2, 0.6, 0.2])
            [Drift('dr', 0.1), Drift('dr', 0.3), Drift('dr', 0.1)]
        """
        # Bx default, the element is indivisible
        return [self]

    def swap_faces(self, copy=False):
        """Swap the faces of an element, alignment errors are ignored"""

        def swapattr(element, attro, attri):
            val = getattr(element, attri)
            delattr(element, attri)
            return attro, val

        if copy:
            el = self.copy()
        else:
            el = self
        # Remove and swap entrance and exit attributes
        fin = dict(swapattr(el, kout, kin)
                   for kin, kout in zip(el._entrance_fields,
                                        el._exit_fields)
                   if kin in vars(el)
                   and kin not in el._no_swap)
        fout = dict(swapattr(el, kin, kout)
                    for kin, kout in zip(el._entrance_fields,
                                         el._exit_fields)
                    if kout in vars(el)
                    and kout not in el._no_swap)
        # Apply swapped entrance and exit attributes
        for key, value in fin.items():
            setattr(el, key, value)
        for key, value in fout.items():
            setattr(el, key, value)
        return el if copy else None

    def update(self, *args, **kwargs):
        """
        update(**kwargs)
        update(mapping, **kwargs)
        update(iterable, **kwargs)
        Update the element attributes with the given arguments
        """
        attrs = dict(*args, **kwargs)
        for key, value in attrs.items():
            setattr(self, key, value)

    def copy(self) -> Element:
        """Return a shallow copy of the element"""
        return copy(self)

    def deepcopy(self) -> Element:
        """Return a deep copy of the element"""
        return deepcopy(self)

    @property
    def definition(self) -> tuple[str, tuple, dict]:
        """tuple (class_name, args, kwargs) defining the element"""
        attrs = dict(self.items())
        arguments = tuple(attrs.pop(
            k, getattr(self, k)) for k in self._BUILD_ATTRIBUTES
        )
        defelem = self.__class__(*arguments)
        keywords = dict(
            (k, v)
            for k, v in attrs.items()
            if not numpy.array_equal(v, getattr(defelem, k, None))
        )
        return self.__class__.__name__, arguments, keywords

    def items(self) -> Generator[tuple[str, Any], None, None]:
        """Iterates through the data members"""
        v = vars(self).copy()
        for k in ["FamName", "Length", "PassMethod"]:
            yield k, v.pop(k)
        for k, v in sorted(v.items()):
            yield k, v

    def is_compatible(self, other: Element) -> bool:
        """Checks if another :py:class:`Element` can be merged"""
        return False

    def merge(self, other) -> None:
        """Merge another element"""
        if not self.is_compatible(other):
            badname = getattr(other, 'FamName', type(other))
            raise TypeError("Cannot merge {0} and {1}".format(self.FamName, badname))

    # noinspection PyMethodMayBeStatic
    def _get_longt_motion(self):
        return False

    # noinspection PyMethodMayBeStatic
    def _get_collective(self):
        return False

    @property
    def longt_motion(self) -> bool:
        """:py:obj:`True` if longitudinal motion is affected by the element"""
        return self._get_longt_motion()

    @property
    def is_collective(self) -> bool:
        """:py:obj:`True` if the element involves collective effects"""
        return self._get_collective()


class LongElement(Element):
    """Base class for long elements"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Length']

    def __init__(self, family_name: str, length: float, *args, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]

        Other arguments and keywords are given to the base class
        """
        kwargs.setdefault('Length', length)
        # Ancestor may be either Element of ThinMultipole
        # noinspection PyArgumentList
        super(LongElement, self).__init__(family_name, *args, **kwargs)

    def _part(self, fr, sumfr):
        pp = self.copy()
        pp.Length = fr * self.Length
        if hasattr(self, 'KickAngle'):
            pp.KickAngle = fr / sumfr * self.KickAngle
        return pp

    def divide(self, frac) -> list[Element]:
        def popattr(element, attr):
            val = getattr(element, attr)
            delattr(element, attr)
            return attr, val

        frac = numpy.asarray(frac, dtype=float)
        el = self.copy()
        # Remove entrance and exit attributes
        fin = dict(popattr(el, key) for key in vars(self) if
                   key in self._entrance_fields)
        fout = dict(popattr(el, key) for key in vars(self) if key in self._exit_fields)
        # Split element
        element_list = [el._part(f, numpy.sum(frac)) for f in frac]
        # Restore entrance and exit attributes
        for key, value in fin.items():
            setattr(element_list[0], key, value)
        for key, value in fout.items():
            setattr(element_list[-1], key, value)
        return element_list

    def is_compatible(self, other) -> bool:
        def compatible_field(fieldname):
            f1 = getattr(self, fieldname, None)
            f2 = getattr(other, fieldname, None)
            if f1 is None and f2 is None:  # no such field
                return True
            elif f1 is None or f2 is None:  # only one
                return False
            else:  # both
                return numpy.all(f1 == f2)

        if not (type(other) is type(self) and self.PassMethod == other.PassMethod):
            return False
        for fname in ("RApertures", "EApertures"):
            if not compatible_field(fname):
                return False
        return True

    def merge(self, other) -> None:
        super().merge(other)
        self.Length += other.Length


class Marker(Element):
    """Marker element"""


class Monitor(Element):
    """Monitor element"""


class BeamMoments(Element):
    """Element to compute bunches mean and std"""

    def __init__(self, family_name: str, **kwargs):
        """
        Args:
            family_name:    Name of the element

        Default PassMethod: ``BeamMomentsPass``
        """
        kwargs.setdefault('PassMethod', 'BeamMomentsPass')
        self._stds = numpy.zeros((6, 1, 1), order='F')
        self._means = numpy.zeros((6, 1, 1), order='F')
        super(BeamMoments, self).__init__(family_name, **kwargs)

    def set_buffers(self, nturns, nbunch):
        self._stds = numpy.zeros((6, nbunch, nturns), order='F')
        self._means = numpy.zeros((6, nbunch, nturns), order='F')

    @property
    def stds(self):
        """Beam 6d standard deviation"""
        return self._stds

    @property
    def means(self):
        """Beam 6d center of mass"""
        return self._means


class SliceMoments(Element):
    """Element to compute slices mean and std"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['nslice']
    _conversions = dict(Element._conversions, nslice=int)

    def __init__(self, family_name: str, nslice: int, **kwargs):
        """
        Args:
            family_name:    Name of the element
            nslice:         Number of slices

        Keyword arguments:
            startturn:      Start turn of the acquisition (Default 0)
            endturn:        End turn of the acquisition (Default 1)

        Default PassMethod: ``SliceMomentsPass``
        """
        kwargs.setdefault('PassMethod', 'SliceMomentsPass')
        self._startturn = kwargs.pop('startturn', 0)
        self._endturn = kwargs.pop('endturn', 1)
        super(SliceMoments, self).__init__(family_name, nslice=nslice, **kwargs)
        self._nbunch = 1
        self.startturn = self._startturn
        self.endturn = self._endturn
        self._dturns = self.endturn - self.startturn
        self._stds = numpy.zeros((3, nslice, self._dturns), order='F')
        self._means = numpy.zeros((3, nslice, self._dturns), order='F')
        self._spos = numpy.zeros((nslice, self._dturns), order='F')
        self._weights = numpy.zeros((nslice, self._dturns), order='F')
        self.set_buffers(self._endturn, 1)

    def set_buffers(self, nturns, nbunch):
        self.endturn = min(self.endturn, nturns)
        self._dturns = self.endturn - self.startturn
        self._nbunch = nbunch
        self._stds = numpy.zeros((3, nbunch*self.nslice, self._dturns), order="F")
        self._means = numpy.zeros((3, nbunch*self.nslice, self._dturns), order="F")
        self._spos = numpy.zeros((nbunch*self.nslice, self._dturns), order="F")
        self._weights = numpy.zeros((nbunch*self.nslice, self._dturns), order="F")

    @property
    def stds(self):
        """Slices x,y,dp standard deviation"""
        return self._stds.reshape((3, self._nbunch, self.nslice, self._dturns))

    @property
    def means(self):
        """Slices x,y,dp center of mass"""
        return self._means.reshape((3, self._nbunch, self.nslice, self._dturns))

    @property
    def spos(self):
        """Slices s position"""
        return self._spos.reshape((self._nbunch, self.nslice, self._dturns))

    @property
    def weights(self):
        """Slices weights in mA if beam current >0,
        otherwise fraction of total number of
        particles in the bunch
        """
        return self._weights.reshape((self._nbunch, self.nslice, self._dturns))

    @property
    def startturn(self):
        """Start turn of the acquisition"""
        return self._startturn

    @startturn.setter
    def startturn(self, value):
        if value < 0:
            raise ValueError('startturn must be greater or equal to 0')
        if value >= self._endturn:
            raise ValueError('startturn must be smaller than endturn')
        self._startturn = value

    @property
    def endturn(self):
        """End turn of the acquisition"""
        return self._endturn

    @endturn.setter
    def endturn(self, value):
        if value <= 0:
            raise ValueError('endturn must be greater than 0')
        if value <= self._startturn:
            raise ValueError('endturn must be greater than startturn')
        self._endturn = value


class Aperture(Element):
    """Aperture element"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Limits']
    _conversions = dict(Element._conversions, Limits=lambda v: _array(v, (4,)))

    def __init__(self, family_name, limits, **kwargs):
        """
        Args:
            family_name:    Name of the element
            limits:         (4,) array of physical aperture:
              [xmin, xmax, zmin, zmax] [m]

        Default PassMethod: ``AperturePass``
        """
        kwargs.setdefault('PassMethod', 'AperturePass')
        super(Aperture, self).__init__(family_name, Limits=limits, **kwargs)


class Drift(LongElement):
    """Drift space element"""

    def __init__(self, family_name: str, length: float, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]

        Default PassMethod: ``DriftPass``
        """
        kwargs.setdefault('PassMethod', 'DriftPass')
        super(Drift, self).__init__(family_name, length, **kwargs)

    def insert(self,
               insert_list: Iterable[tuple[float, Optional[Element]]]) \
            -> list[Element]:
        """insert elements inside a drift

        Arguments:
            insert_list: iterable, each item of insert_list is itself an
              iterable with 2 objects:

              1. the location where the center of the element
                 will be inserted, given as a fraction of the Drift length.
              2. an element to be inserted at that location. If :py:obj:`None`,
                 the drift will be divided but no element will be inserted.

        Returns:
             elem_list: a list of elements.

        Drifts with negative lengths may be generated if necessary.

        Examples:

            >>> Drift('dr', 2.0).insert(((0.25, None), (0.75, None)))
            [Drift('dr', 0.5), Drift('dr', 1.0), Drift('dr', 0.5)]

            >>> Drift('dr', 2.0).insert(((0.0, Marker('m1')),
            ... (0.5, Marker('m2'))))
            [Marker('m1'), Drift('dr', 1.0), Marker('m2'), Drift('dr', 1.0)]

            >>> Drift('dr', 2.0).insert(((0.5, Quadrupole('qp', 0.4, 0.0)),))
            [Drift('dr', 0.8), Quadrupole('qp', 0.4), Drift('dr', 0.8)]
        """
        frac, elements = zip(*insert_list)
        lg = [0.0 if el is None else el.Length for el in elements]
        fr = numpy.asarray(frac, dtype=float)
        lg = 0.5 * numpy.asarray(lg, dtype=float) / self.Length
        drfrac = numpy.hstack((fr - lg, 1.0)) - numpy.hstack((0.0, fr + lg))
        long_elems = (drfrac != 0.0)
        drifts = numpy.ndarray((len(drfrac),), dtype='O')
        drifts[long_elems] = self.divide(drfrac[long_elems])
        nline = len(drifts) + len(elements)
        line = [None] * nline  # type: list[Optional[Element]]
        line[::2] = drifts
        line[1::2] = elements
        return [el for el in line if el is not None]


class Collimator(Drift):
    """Collimator element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['RApertures']

    def __init__(self, family_name: str, length: float, limits, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            limits:         (4,) array of physical aperture:
              [xmin, xmax, zmin, zmax] [m]

        Default PassMethod: ``DriftPass``
        """
        super(Collimator, self).__init__(family_name, length,
                                         RApertures=limits, **kwargs)


class ThinMultipole(Element):
    """Thin multipole element"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ["PolynomA", "PolynomB"]

    def __init__(self, family_name: str, poly_a, poly_b, **kwargs):
        """
        Args:
            family_name:    Name of the element
            poly_a:         Array of skew multipole components
            poly_b:         Array of normal multipole components

        Keyword arguments:
            MaxOrder:       Number of desired multipoles. Default: highest
              index of non-zero polynomial coefficients
            FieldScaling:   Scaling factor applied to the magnetic field
              (*PolynomA* and *PolynomB*)

        Default PassMethod: ``ThinMPolePass``
        """

        def getpol(poly):
            nonzero = numpy.flatnonzero(poly != 0.0)
            return poly, len(poly), nonzero[-1] if len(nonzero) > 0 else -1

        def lengthen(poly, dl):
            if dl > 0:
                return numpy.concatenate((poly, numpy.zeros(dl)))
            else:
                return poly

        # PolynomA and PolynomB and convert to ParamArray
        prmpola = self._conversions["PolynomA"](kwargs.pop("PolynomA", poly_a))
        prmpolb = self._conversions["PolynomB"](kwargs.pop("PolynomB", poly_b))
        poly_a, len_a, ord_a = getpol(prmpola)
        poly_b, len_b, ord_b = getpol(prmpolb)
        deforder = max(getattr(self, 'DefaultOrder', 0), ord_a, ord_b)
        # Remove MaxOrder
        maxorder = kwargs.pop('MaxOrder', deforder)
        kwargs.setdefault('PassMethod', 'ThinMPolePass')
        super(ThinMultipole, self).__init__(family_name, **kwargs)
        # Set MaxOrder while PolynomA and PolynomB are not set yet
        super(ThinMultipole, self).__setattr__('MaxOrder', maxorder)
        # Adjust polynom lengths and set them
        len_ab = max(self.MaxOrder + 1, len_a, len_b)
        self.PolynomA = lengthen(prmpola, len_ab - len_a)
        self.PolynomB = lengthen(prmpolb, len_ab - len_b)

    def __setattr__(self, key, value):
        """Check the compatibility of MaxOrder, PolynomA and PolynomB"""
        polys = ('PolynomA', 'PolynomB')
        if key in polys:
            lmin = self.MaxOrder
            if not len(value) > lmin:
                raise ValueError(
                    'Length of {0} must be larger than {1}'.format(key, lmin))
        elif key == 'MaxOrder':
            intval = int(value)
            lmax = min(len(getattr(self, k)) for k in polys)
            if not intval < lmax:
                raise ValueError("MaxOrder must be smaller than {0}".format(lmax))
        super(ThinMultipole, self).__setattr__(key, value)


class Multipole(_Radiative, LongElement, ThinMultipole):
    """Multipole element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ["PolynomA", "PolynomB"]
    _conversions = dict(ThinMultipole._conversions, K=float, H=float)

    def __init__(self, family_name: str, length: float, poly_a, poly_b, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            poly_a:         Array of skew multipole components
            poly_b:         Array of normal multipole components

        Keyword arguments:
            MaxOrder:       Number of desired multipoles. Default: highest
              index of non-zero polynomial coefficients
            NumIntSteps:    Number of integration steps (default: 10)
            KickAngle:      Correction deviation angles (H, V)
            FieldScaling:   Scaling factor applied to the magnetic field
              (*PolynomA* and *PolynomB*)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        kwargs.setdefault('NumIntSteps', 10)
        super(Multipole, self).__init__(family_name, length, poly_a, poly_b, **kwargs)

    def is_compatible(self, other) -> bool:
        if super().is_compatible(other) and self.MaxOrder == other.MaxOrder:
            for i in range(self.MaxOrder + 1):
                if self.PolynomB[i] != other.PolynomB[i]:
                    return False
                if self.PolynomA[i] != other.PolynomA[i]:
                    return False
            return True
        else:
            return False

    # noinspection PyPep8Naming
    @property
    def K(self) -> float:
        """Focusing strength [mˆ-2]"""
        arr = self.PolynomB
        return 0.0 if len(arr) < 2 else arr[1]

    # noinspection PyPep8Naming
    @K.setter
    def K(self, strength: float):
        self.PolynomB[1] = strength

    # noinspection PyPep8Naming
    @property
    def H(self) -> float:
        """Sextupolar strength [mˆ-3]"""
        arr = self.PolynomB
        return 0.0 if len(arr) < 3 else arr[2]

    # noinspection PyPep8Naming
    @H.setter
    def H(self, strength):
        self.PolynomB[2] = strength


class Dipole(Radiative, Multipole):
    """Dipole element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['BendingAngle',
                                                         'K']
    _conversions = dict(Multipole._conversions, EntranceAngle=float,
                        ExitAngle=float,
                        FringeInt1=float, FringeInt2=float,
                        FringeQuadEntrance=int, FringeQuadExit=int,
                        FringeBendEntrance=int, FringeBendExit=int)

    _entrance_fields = Multipole._entrance_fields + ['EntranceAngle',
                                                     'FringeInt1',
                                                     'FringeBendEntrance',
                                                     'FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['ExitAngle',
                                             'FringeInt2',
                                             'FringeBendExit',
                                             'FringeQuadExit']

    DefaultOrder = 0

    def __init__(self, family_name: str, length: float,
                 bending_angle: Optional[float] = 0.0, k: float = 0.0,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            bending_angle:  Bending angle [rd]
            k:              Focusing strength [m^-2]

        Keyword arguments:
            EntranceAngle=0.0:  entrance angle
            ExitAngle=0.0:      exit angle
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder=0:         Number of desired multipoles
            NumIntSt=10:        Number of integration steps
            FullGap:            Magnet full gap
            FringeInt1:         Extension of the entrance fringe field
            FringeInt2:         Extension of the exit fringe field
            FringeBendEntrance: 1: legacy version Brown First Order (default)

              2: SOLEIL close to second order of Brown

              3: THOMX
            FringeBendExit:     See *FringeBendEntrance*
            FringeQuadEntrance: 0: no fringe field effect (default)

              1: Lee-Whiting's thin lens limit formula

              2: elegant-like
            FringeQuadExit:     See *FringeQuadEntrance*
            fringeIntM0:        Integrals for FringeQuad method 2
            fringeIntP0:
            KickAngle:          Correction deviation angles (H, V)
            FieldScaling:       Scaling factor applied to the magnetic field

        Available PassMethods: :ref:`BndMPoleSymplectic4Pass`,
        :ref:`BendLinearPass`, :ref:`ExactSectorBendPass`,
        :ref:`ExactRectangularBendPass`, :ref:`ExactRectBendPass`,
        BndStrMPoleSymplectic4Pass

        Default PassMethod: :ref:`BndMPoleSymplectic4Pass`
        """
        kwargs.setdefault('BendingAngle', bending_angle)
        kwargs.setdefault('EntranceAngle', 0.0)
        kwargs.setdefault('ExitAngle', 0.0)
        kwargs.setdefault('PassMethod', 'BndMPoleSymplectic4Pass')
        super(Dipole, self).__init__(family_name, length, [], [0.0, k], **kwargs)

    def items(self) -> Generator[tuple[str, Any], None, None]:
        yield from super().items()
        yield "K", vars(self)["PolynomB"][1]

    def _part(self, fr, sumfr):
        pp = super(Dipole, self)._part(fr, sumfr)
        pp.BendingAngle = fr / sumfr * self.BendingAngle
        pp.EntranceAngle = 0.0
        pp.ExitAngle = 0.0
        return pp

    def is_compatible(self, other) -> bool:
        def invrho(dip: Dipole):
            return dip.BendingAngle / dip.Length

        return (super().is_compatible(other)
                and self.ExitAngle == -other.EntranceAngle
                and abs(invrho(self) - invrho(other)) <= 1.e-6)

    def merge(self, other) -> None:
        super().merge(other)
        # noinspection PyAttributeOutsideInit
        self.ExitAngle = other.ExitAngle
        self.BendingAngle += other.BendingAngle


# Bend is a synonym of Dipole.
Bend = Dipole


class Quadrupole(Radiative, Multipole):
    """Quadrupole element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['K']
    _conversions = dict(Multipole._conversions, FringeQuadEntrance=int,
                        FringeQuadExit=int)

    _entrance_fields = Multipole._entrance_fields + ['FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['FringeQuadExit']

    DefaultOrder = 1

    def __init__(self, family_name: str, length: float,
                 k: Optional[float] = 0.0, **kwargs):
        """Quadrupole(FamName, Length, Strength=0, **keywords)

        Args:
            family_name:    Name of the element
            length:         Element length [m]
            k:              Focusing strength [mˆ-2]

        Keyword Arguments:
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder=1:         Number of desired multipoles
            NumIntSteps=10:     Number of integration steps
            FringeQuadEntrance: 0: no fringe field effect (default)

              1: Lee-Whiting's thin lens limit formula

              2: elegant-like
            FringeQuadExit:     See ``FringeQuadEntrance``
            fringeIntM0:        Integrals for FringeQuad method 2
            fringeIntP0:
            KickAngle:          Correction deviation angles (H, V)
            FieldScaling:       Scaling factor applied to the magnetic field
              (*PolynomA* and *PolynomB*)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        kwargs.setdefault("PassMethod", "StrMPoleSymplectic4Pass")
        super(Quadrupole, self).__init__(family_name, length, [], [0.0, k], **kwargs)

    def items(self) -> Generator[tuple[str, Any], None, None]:
        yield from super().items()
        yield "K", vars(self)["PolynomB"][1]


class Sextupole(Multipole):
    """Sextupole element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['H']

    DefaultOrder = 2

    def __init__(self, family_name: str, length: float,
                 h: Optional[float] = 0.0, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            h:              strength [mˆ-3]

        Keyword Arguments:
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder:           Number of desired multipoles
            NumIntSteps=10:     Number of integration steps
            KickAngle:          Correction deviation angles (H, V)
            FieldScaling:       Scaling factor applied to the magnetic field
              (*PolynomA* and *PolynomB*)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        kwargs.setdefault("PassMethod", "StrMPoleSymplectic4Pass")
        super(Sextupole, self).__init__(family_name, length, [], [0.0, 0.0, h],
                                        **kwargs)

    def items(self) -> Generator[tuple[str, Any], None, None]:
        yield from super().items()
        yield "H", vars(self)["PolynomB"][2]


class Octupole(Multipole):
    """Octupole element, with no changes from multipole at present"""

    _BUILD_ATTRIBUTES = Multipole._BUILD_ATTRIBUTES

    DefaultOrder = 3


class RFCavity(LongtMotion, LongElement):
    """RF cavity element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['Voltage',
                                                         'Frequency',
                                                         'HarmNumber',
                                                         'Energy']
    default_pass = {False: 'DriftPass', True: 'RFCavityPass'}
    _conversions = dict(LongElement._conversions,
                        Voltage=float, Frequency=float,
                        HarmNumber=int, TimeLag=float)

    def __init__(self, family_name: str, length: float, voltage: float,
                 frequency: float, harmonic_number: int, energy: float,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            voltage:        RF voltage [V]
            frequency:      RF frequency [Hz]
            harmonic_number:
            energy:         ring energy [eV]

        Keyword Arguments:
            TimeLag=0:      Cavity time lag

        Default PassMethod: ``RFCavityPass``
        """
        kwargs.setdefault('TimeLag', 0.0)
        kwargs.setdefault('PassMethod', self.default_pass[True])
        super(RFCavity, self).__init__(family_name, length,
                                       Voltage=voltage,
                                       Frequency=frequency,
                                       HarmNumber=harmonic_number,
                                       Energy=energy, **kwargs)

    def _part(self, fr, sumfr):
        pp = super(RFCavity, self)._part(fr, sumfr)
        pp.Voltage = fr * self.Voltage
        return pp

    def is_compatible(self, other) -> bool:
        return (super().is_compatible(other)
                and self.Frequency == other.Frequency
                and self.TimeLag == other.TimeLag)

    def merge(self, other) -> None:
        super().merge(other)
        self.Voltage += other.Voltage

    def _get_longt_motion(self):
        return self.PassMethod.endswith('CavityPass')

    # noinspection PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, **kwargs):
        if new_pass == 'auto':
            new_pass = (self.default_pass[True] if enable else
                        ('IdentityPass' if self.Length == 0 else 'DriftPass'))
        return super().set_longt_motion(enable, new_pass=new_pass, **kwargs)


class M66(Element):
    """Linear (6, 6) transfer matrix"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ["M66"]
    _conversions = dict(Element._conversions, M66=_array66)

    def __init__(self, family_name: str, m66=None, **kwargs):
        """
        Args:
            family_name:    Name of the element
            m66:            Transfer matrix. Default: Identity matrix

        Default PassMethod: ``Matrix66Pass``
        """
        if m66 is None:
            m66 = numpy.identity(6)
        kwargs.setdefault('PassMethod', 'Matrix66Pass')
        kwargs.setdefault("M66", m66)
        super(M66, self).__init__(family_name, **kwargs)


class SimpleQuantDiff(_DictLongtMotion, Element):
    """
    Linear tracking element for a simplified quantum diffusion,
    radiation damping and energy loss.

    Note: The damping times are needed to compute the correct
    kick for the emittance. Radiation damping is NOT applied.
    """

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES
    default_pass = {False: 'IdentityPass', True: 'SimpleQuantDiffPass'}

    def __init__(self, family_name: str, betax: float = 1.0,
                 betay: float = 1.0, emitx: float = 0.0,
                 emity: float = 0.0, espread: float = 0.0,
                 taux: float = 0.0, tauy: float = 0.0,
                 tauz: float = 0.0,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element

        Optional Args:
            betax:         Horizontal beta function at element [m]
            betay:         Vertical beta function at element [m]
            emitx:         Horizontal equilibrium emittance [m.rad]
            emity:         Vertical equilibrium emittance [m.rad]
            espread:       Equilibrium energy spread
            taux:          Horizontal damping time [turns]
            tauy:          Vertical damping time [turns]
            tauz:          Longitudinal damping time [turns]

        Default PassMethod: ``SimpleQuantDiffPass``
        """
        kwargs.setdefault("PassMethod", self.default_pass[True])

        assert taux >= 0.0, 'taux must be greater than or equal to 0'
        self.taux = taux

        assert tauy >= 0.0, 'tauy must be greater than or equal to 0'
        self.tauy = tauy

        assert tauz >= 0.0, 'tauz must be greater than or equal to 0'
        self.tauz = tauz

        assert emitx >= 0.0, 'emitx must be greater than or equal to 0'
        self.emitx = emitx
        if emitx > 0.0:
            assert taux > 0.0, 'if emitx is given, taux must be non zero'

        assert emity >= 0.0, 'emity must be greater than or equal to 0'
        self.emity = emity
        if emity > 0.0:
            assert tauy > 0.0, 'if emity is given, tauy must be non zero'

        assert espread >= 0.0, 'espread must be greater than or equal to 0'
        self.espread = espread
        if espread > 0.0:
            assert tauz > 0.0, 'if espread is given, tauz must be non zero'

        self.betax = betax
        self.betay = betay
        super(SimpleQuantDiff, self).__init__(family_name, **kwargs)


class SimpleRadiation(_DictLongtMotion, Radiative, Element):
    """Simple radiation damping and energy loss"""

    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES
    _conversions = dict(Element._conversions, U0=float,
                        damp_mat_diag=lambda v: _array(v, shape=(6,)))

    default_pass = {False: 'IdentityPass', True: 'SimpleRadiationPass'}

    def __init__(self, family_name: str,
                 taux: float = 0.0, tauy: float = 0.0,
                 tauz: float = 0.0, U0: float = 0.0,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element

        Optional Args:
            taux:          Horizontal damping time [turns]
            tauy:          Vertical damping time [turns]
            tauz:          Longitudinal damping time [turns]
            U0:            Energy loss per turn [eV]

        Default PassMethod: ``SimpleRadiationPass``
        """
        assert taux >= 0.0, 'taux must be greater than or equal to 0'
        if taux == 0.0:
            dampx = 1
        else:
            dampx = numpy.exp(-1/taux)

        assert tauy >= 0.0, 'tauy must be greater than or equal to 0'
        if tauy == 0.0:
            dampy = 1
        else:
            dampy = numpy.exp(-1/tauy)

        assert tauz >= 0.0, 'tauz must be greater than or equal to 0'
        if tauz == 0.0:
            dampz = 1
        else:
            dampz = numpy.exp(-1/tauz)

        kwargs.setdefault('PassMethod', self.default_pass[True])
        kwargs.setdefault("U0", U0)
        kwargs.setdefault("damp_mat_diag",
                          numpy.array([dampx, dampx, dampy, dampy, dampz, dampz]))

        super(SimpleRadiation, self).__init__(family_name, **kwargs)


class Corrector(LongElement):
    """Corrector element"""

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['KickAngle']

    def __init__(self, family_name: str, length: float, kick_angle, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            KickAngle:      Correction deviation angles (H, V)

        Keyword Args:
            FieldScaling:   Scaling factor applied to the magnetic field
              (*KickAngle*)

        Default PassMethod: ``CorrectorPass``
        """
        kwargs.setdefault('PassMethod', 'CorrectorPass')
        super(Corrector, self).__init__(family_name, length,
                                        KickAngle=kick_angle, **kwargs)


class Wiggler(Radiative, LongElement):
    """Wiggler element

    See atwiggler.m
    """

    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['Lw', 'Bmax',
                                                         'Energy']
    _conversions = dict(Element._conversions, Lw=float, Bmax=float,
                        Energy=float,
                        Bx=lambda v: _array(v, (6, -1)),
                        By=lambda v: _array(v, (6, -1)),
                        Nstep=int, Nmeth=int, NHharm=int, NVharm=int)

    # noinspection PyPep8Naming
    def __init__(self, family_name: str, length: float, wiggle_period: float,
                 b_max: float, energy: float, Nstep: Optional[int] = 5,
                 Nmeth: Optional[int] = 4,
                 By=(1, 1, 0, 1, 1, 0), Bx=(), **kwargs):
        """
        Args:
            length:         total length of the wiggler
            wiggle_period:  length must be a multiple of this
            b_max:          peak wiggler field [Tesla]
            energy:         beam energy [eV]
            Nstep:          number of integration steps.
            Nmeth:          symplectic integration order: 2 or 4
            Bx:             harmonics for horizontal wiggler: (6, nHharm)
                              array-like object
            By:             harmonics for vertical wiggler (6, nHharm)
                              array-like object

        Default PassMethod: ``GWigSymplecticPass``
        """
        kwargs.setdefault('PassMethod', 'GWigSymplecticPass')
        n_wiggles = length / wiggle_period
        if abs(round(n_wiggles) - n_wiggles) > 1e-6:
            raise ValueError("Wiggler: length / wiggle_period is not an "
                             "integer. ({0}/{1}={2})".format(length,
                                                             wiggle_period,
                                                             n_wiggles))
        super(Wiggler, self).__init__(family_name, length, Lw=wiggle_period,
                                      Bmax=b_max, Nstep=Nstep, Nmeth=Nmeth,
                                      By=By, Bx=Bx, Energy=energy, **kwargs)

        for i, b in enumerate(self.By.T):
            dk = abs(b[3] ** 2 - b[4] ** 2 - b[2] ** 2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(H): kx^2 + kz^2 -ky^2 !=0, i = {0}".format(i))

        for i, b in enumerate(self.Bx.T):
            dk = abs(b[2] ** 2 - b[4] ** 2 - b[3] ** 2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(V): ky^2 + kz^2 -kx^2 !=0, i = {0}".format(i))

        self.NHharm = self.By.shape[1]
        self.NVharm = self.Bx.shape[1]


class QuantumDiffusion(_DictLongtMotion, Element):
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Lmatp']
    default_pass = {False: 'IdentityPass', True: 'QuantDiffPass'}
    _conversions = dict(Element._conversions, Lmatp=_array66)

    def __init__(self, family_name: str, lmatp: numpy.ndarray, **kwargs):
        """Quantum diffusion element

        Args:
            family_name:    Name of the element
            lmatp      :    Diffusion matrix for generation (see
              :py:func:`.gen_quantdiff_elem`)

        Default PassMethod: ``QuantDiffPass``
        """
        kwargs.setdefault('PassMethod', self.default_pass[True])
        super().__init__(family_name, Lmatp=lmatp, **kwargs)


class EnergyLoss(_DictLongtMotion, Element):
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['EnergyLoss']
    _conversions = dict(Element._conversions, EnergyLoss=float)
    default_pass = {False: 'IdentityPass', True: 'EnergyLossRadPass'}

    def __init__(self, family_name: str, energy_loss: float, **kwargs):
        """Energy loss element

        Args:
            family_name:    Name of the element
            energy_loss:    Energy loss [eV]

        """
        kwargs.setdefault('PassMethod', self.default_pass[False])
        super().__init__(family_name, EnergyLoss=energy_loss, **kwargs)


Radiative.register(EnergyLoss)


def build_class_map():  # Missing class aliases (Bend)
    global CLASS_MAP

    def subclasses_recursive(cl):
        direct = cl.__subclasses__()
        indirect = []
        for subclass in direct:
            indirect.extend(subclasses_recursive(subclass))
        return frozenset([cl] + direct + indirect)

    cls_list = subclasses_recursive(Element)
    CLASS_MAP = {cls.__name__: cls for cls in cls_list}


def get_class_map():
    return CLASS_MAP


# build_class_map()

CLASS_MAP = dict((k, v) for k, v in locals().items()
                 if isinstance(v, type) and issubclass(v, Element))



"""
Module to compute the touschek lifetime of the ring
"""

import at
import numpy
from ..lattice import Lattice, AtError, AtWarning
import warnings
from scipy.special import iv
from scipy import integrate
from scipy.optimize import fsolve
from ..constants import qe, clight, _e_radius


__all__ = ['get_bunch_length_espread', 'get_lifetime', 'get_scattering_rate']


def get_bunch_length_espread(ring, zn=None, bunch_curr=None, espread=None):
    """Haissinski equation solver

    Solves the Haissinski formula and returns the bunch length and energy
    spread for given bunch current and :math:`Z/n`. If both ``zn`` and
    ``bunch_curr`` are ``None``, zero current case, otherwise both are needed
    for the calculation

    args:
        ring:             ring use for tracking

    keyword args:
        zn=None:          :math:`Z/n` for the full ring
        bunch_curr=None:  Bunch current
        espread=None:     Energy spread, if ``None`` use lattice parameter

    Returns:
        Bunch length, energy spread
    """
    def haissinski(x, cst):
        return x**3-x-cst

    ep = at.envelope_parameters(ring.radiation_on(copy=True))
    bl0 = ep.sigma_l
    if espread is None:
        espread = ep.sigma_e
    if zn is None and bunch_curr is None:
        bl = bl0
    elif zn is None or bunch_curr is None:
        raise AtError('Please provide both current and Z/n for bunch '
                      'length calculation')
    else:
        vrf = ring.get_rf_voltage()
        h = ring.harmonic_number
        etac = at.get_slip_factor(ring.radiation_off(copy=True))
        cs = numpy.cos(ep.phi_s)
        nus = ep.f_s/ring.revolution_frequency
        cst = (-0.5 * numpy.sqrt(numpy.pi) * bunch_curr * zn /
               (vrf * h * cs * (abs(etac) * espread / nus)**3))
        bl = bl0*fsolve(haissinski, numpy.array([1.0]), args=cst)[0]
    return bl, espread


def get_beam_sizes(ring, bunch_curr, zn=None, emitx=None,
                   sigs=None, sigp=None):
    if zn is None:
        bunch_curr = None
    if sigs is None or sigp is None:
        sigsi, sigpi = get_bunch_length_espread(ring, zn=zn,
                                                bunch_curr=bunch_curr,
                                                espread=sigp)
        if sigs is None:
            sigs = sigsi
        if sigp is None:
            sigp = sigpi
    if emitx is None:
        _, bd, _ = at.ohmi_envelope(ring.radiation_on(copy=True))
        emitx = bd.mode_emittances[0]
    return emitx, sigs, sigp


def int_piwinski(k, km, B1, B2):
    r"""
    Integrand of the piwinski formula
    In case the Bessel function has too large value
    (more than :math:`10^251`) it
    is substituted by its exponential approximation:
    :math:`I_0(x)~\frac{\exp(x)}{\sqrt{2 \pi x}}`
    """
    t = numpy.tan(k)**2
    tm = numpy.tan(km)**2
    fact = ((2*t+1)**2*(t/tm/(1+t)-1)/t + t - numpy.sqrt(t*tm*(1+t)) -
            (2+1/(2*t))*numpy.log(t/tm/(1+t)))
    if B2*t < 500:
        intp = fact * numpy.exp(-B1*t)*iv(0, B2*t)*numpy.sqrt(1+t)
    else:
        intp = fact * numpy.exp(B2*t-B1*t) / \
             numpy.sqrt(2*numpy.pi*B2*t)*numpy.sqrt(1+t)
    return intp


def _get_vals(ring, rp, ma, emity, bunch_curr, emitx=None,
              sigs=None, sigp=None, zn=None, epsabs=1.0e-16,
              epsrel=1.0e-12):

    emitx, sigs, sigp = get_beam_sizes(ring, bunch_curr, zn=zn,
                                       emitx=emitx, sigs=sigs,
                                       sigp=sigp)

    nc = bunch_curr/ring.revolution_frequency/qe
    beta2 = ring.beta*ring.beta
    gamma2 = ring.gamma*ring.gamma

    emit = numpy.array([emitx, emity])
    _, _, ld = ring.get_optics(refpts=rp)
    bxy = ld.beta
    bxy2 = bxy*bxy
    axy = ld.alpha
    dxy = ld.dispersion[:, [0, 2]]
    dxy2 = dxy*dxy
    dpxy = ld.dispersion[:, [1, 3]]
    sigb = numpy.sqrt(bxy*emit)
    sigb2 = sigb*sigb
    sigp2 = sigp*sigp
    sig = numpy.sqrt(sigb2+sigp2*dxy2)
    sig2 = sig*sig
    dt = dxy*axy+dpxy*bxy
    dt2 = dt*dt
    sigh2 = 1/(1/sigp2 + numpy.sum((dxy2+dt2)/sigb2, axis=1))

    bs = bxy2/sigb2*(1-(sigh2*(dt2/sigb2).T).T)
    bg2i = 1/(2*beta2*gamma2)
    B1 = bg2i*numpy.sum(bs, axis=1)
    B2sq = bg2i*bg2i*(numpy.diff(bs, axis=1).T**2 +
                      4*sigh2*sigh2*numpy.prod(bxy2*dt2, axis=1) /
                      numpy.prod(sigb2*sigb2, axis=1))
    B2 = numpy.squeeze(numpy.sqrt(B2sq))

    val = numpy.zeros((2, len(rp)))
    for i in range(2):
        dpp = ma[:, i]
        um = beta2*dpp*dpp
        km = numpy.arctan(numpy.sqrt(um))

        for ii in range(len(rp)):
            args = (km[ii], B1[ii], B2[ii])
            val[i, ii], *_ = integrate.quad(int_piwinski, args[0], numpy.pi/2,
                                            args=args, epsabs=epsabs,
                                            epsrel=epsrel)

        val[i] *= (_e_radius**2*clight*nc /
                   (8*numpy.pi*gamma2*sigs *
                    numpy.sqrt(numpy.prod(sig2, axis=1) -
                               sigp2*sigp2*numpy.prod(dxy2, axis=1))) *
                   2*numpy.sqrt(numpy.pi*(B1*B1-B2*B2)))
    return val


def _init_ma_rp(ring, refpts=None, offset=None, momap=None,
                interpolate=True, check_zero=True, **kwargs):
    if refpts is None:
        refpts = ring.get_uint32_index(at.All, endpoint=False)
    else:
        refpts = ring.get_uint32_index(refpts)

    if offset is not None:
        try:
            offset = numpy.broadcast_to(offset, (len(refpts), 6))
        except ValueError:
            msg = ('offset and refpts have incoherent '
                   'shapes: {0}, {1}'.format(offset.shape, refpts.shape))
            raise AtError(msg)

    if momap is not None:
        assert len(momap) == len(refpts), \
            'Input momap and refpts have different lengths'

    if check_zero:
        mask = [ring[r].Length > 0.0 for r in refpts]
    else:
        mask = [True for _ in refpts]

    if not numpy.all(mask):
        zerolength_warning = ('zero-length elements removed '
                              'from lifetime calculation')
        warnings.warn(AtWarning(zerolength_warning))

    refpts = refpts[mask]
    if offset is not None:
        offset = offset[mask]

    if momap is None:
        resolution = kwargs.pop('resolution', 1.0e-3)
        amplitude = kwargs.pop('amplitude', 0.1)
        kwargs.update({'refpts': refpts})
        kwargs.update({'offset': offset})
        momap, _, _ = ring.get_momentum_acceptance(resolution,
                                                   amplitude, **kwargs)
    else:
        momap = momap[mask]

    if interpolate:
        refpts_all = numpy.array([i for i in range(refpts[0],
                                                   refpts[-1]+1)
                                  if ring[i].Length > 0])
        spos = numpy.squeeze(ring.get_s_pos(refpts))
        spos_all = numpy.squeeze(ring.get_s_pos(refpts_all))
        momp = numpy.interp(spos_all, spos, momap[:, 0])
        momn = numpy.interp(spos_all, spos, momap[:, 1])
        momap_all = numpy.vstack((momp, momn)).T
        ma, rp = momap_all, refpts_all
    else:
        ma, rp = momap, refpts
    return ma, rp


def get_lifetime(ring, emity, bunch_curr, emitx=None, sigs=None, sigp=None,
                 zn=None, momap=None, refpts=None, offset=None, **kwargs):
    """Touschek lifetime calculation

    Computes the touschek lifetime using the Piwinski formula

    args:
        ring:            ring use for tracking
        emity:           verticla emittance
        bunch_curr:      bunch current

    keyword args:
        emitx=None:      horizontal emittance
        sigs=None:       rms bunch length
        sigp=None:       energy spread
        zn=None:         full ring :math:`Z/n`
        momap=None:      momentum aperture, has to be consistent with
                         ``refpts`` if provided the momentum aperture is
                         not calculated
        refpts=None:     ``refpts`` where the momentum aperture is calculated,
                         the default is to compute it for all elements in the
                         ring, ``len(refpts)>2`` is required
        resolution:      minimum distance between 2 grid points, default=1.0e-3
        amplitude:       max. amplitude for ``RADIAL`` and ``CARTESIAN`` or
                         initial step in ``RECURSIVE``
                         default = 0.1
        nturns=1024:     Number of turns for the tracking
        dp=None:         static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:       ``at.GridMode.CARTESIAN/RADIAL`` track full vector
                         (default). ``at.GridMode.RECURSIVE``: recursive search
        use_mp=False:    Use python multiprocessing (``patpass``, default use
                         ``lattice_pass``). In case multi-processing is not
                         enabled ``GridMode`` is forced to
                         ``RECURSIVE`` (most efficient in single core)
        verbose=False:   Print out some inform
        epsabs, epsrel:  integral absolute and relative tolerances

    Returns:
        tl: touschek lifetime, double expressed in seconds
        ma: momentum aperture (len(refpts), 2) array
        refpts: refpts used for momentum aperture calculation
                (len(refpts), ) array
    """
    interpolate = kwargs.pop('interpolate', True)
    epsabs = kwargs.pop('epsabs', 1.0e-16)
    epsrel = kwargs.pop('epsrel', 1.0e-12)
    ma, rp = _init_ma_rp(ring, refpts=refpts, offset=offset,
                         momap=momap, interpolate=interpolate,
                         **kwargs)
    length_all = numpy.array([e.Length for e in ring[rp]])
    vals = _get_vals(ring, rp, ma, emity, bunch_curr, emitx=emitx,
                     sigs=sigs, sigp=sigp, zn=zn, epsabs=epsabs,
                     epsrel=epsrel)
    tl = 1/numpy.mean([sum(v*length_all.T)/sum(length_all) for v in vals])
    return tl, ma, rp


def get_scattering_rate(ring, emity, bunch_curr, emitx=None, sigs=None,
                        sigp=None, zn=None, momap=None, refpts=None,
                        offset=None, **kwargs):
    """Touschek scattering rate calculation

    Computes the touschek scattering using the Piwinski formula

    args:
        ring:            ring use for tracking
        emity:           verticla emittance
        bunch_curr:      bunch current

    keyword args:
        emitx=None:      horizontal emittance
        sigs=None:       rms bunch length
        sigp=None:       energy spread
        zn=None:         full ring :math:`Z/n`
        momap=None:      momentum aperture, has to be consistent with
                         ``refpts`` if provided the momentum aperture
                         is not calculated
        refpts=None:     ``refpts`` where the momentum aperture is calculated,
                         the default is to compute it for all elements in the
                         ring, ``len(refpts)>2`` is required
        resolution:      minimum distance between 2 grid points, default=1.0e-3
        amplitude:       max. amplitude for ``RADIAL`` and ``CARTESIAN`` or
                         initial step in ``RECURSIVE``
                         default = 0.1
        nturns=1024:     Number of turns for the tracking
        dp=None:         static momentum offset
        grid_mode:       ``at.GridMode.CARTESIAN/RADIAL`` track full vector
                         (default). ``at.GridMode.RECURSIVE``: recursive search
        use_mp=False:    Use python multiprocessing (``patpass``, default use
                         ``lattice_pass``). In case multi-processing is not
                         enabled ``GridMode`` is forced to
                         ``RECURSIVE`` (most efficient in single core)
        verbose=False:   Print out some inform
        epsabs, epsrel:  integral absolute and relative tolerances

    Returns:
        scattering_rate: scattering rate double array (len(refpts), )
                         expressed in event/seconds
        ma: momentum aperture (len(refpts), 2) array
        refpts: refpts used for momentum aperture calculation
                (len(refpts), ) array
    """
    interpolate = kwargs.pop('interpolate', True)
    epsabs = kwargs.pop('epsabs', 1.0e-16)
    epsrel = kwargs.pop('epsrel', 1.0e-12)
    ma, rp = _init_ma_rp(ring, refpts=refpts, momap=momap,
                         interpolate=interpolate, check_zero=False,
                         offset=offset, **kwargs)
    vals = _get_vals(ring, rp, ma, emity, bunch_curr, emitx=emitx,
                     sigs=sigs, sigp=sigp, zn=zn, epsabs=epsabs,
                     epsrel=epsrel)
    scattering_rate = (numpy.mean(vals, axis=0)*bunch_curr /
                       ring.revolution_frequency / qe)
    return scattering_rate, ma, rp


Lattice.get_bunch_length_espread = get_bunch_length_espread
Lattice.get_lifetime = get_lifetime
Lattice.get_scattering_rate = get_scattering_rate



"""
Analytical wake functions
"""
import numpy
from at.constants import clight
from scipy.interpolate import interp1d


def convolve_wakefun(srange, w, sigs, gauss_sigma=10):
    """Convolution of a wake function with a pulse of rms
    length sigs, this is use to generate a wake potential
    that can be added to the output of EM code like GDFIDL"""

    def _gauss(s):
        ampl = 1 / (numpy.sqrt(2 * numpy.pi) * sigs)
        expon = numpy.exp(-s**2 / (2 * sigs**2))
        return ampl * expon

    ds = numpy.diff(srange)[-1]
    s_gauss = numpy.arange(-gauss_sigma*sigs, gauss_sigma*sigs+1e-15, ds)
    gauss = _gauss(s_gauss)

    conv = numpy.convolve(gauss, w, mode='full') * ds
    s_offset = gauss_sigma * sigs - numpy.amin(srange)
    s_conv = numpy.arange(len(conv)) * ds - s_offset

    ifun = interp1d(s_conv, conv, bounds_error=False, fill_value=0)
    conv_wf = ifun(srange)
    return conv_wf


def long_resonator_wf(srange, frequency, qfactor, rshunt, beta):
    """
    
    Define the wake function (longitudinal) of a resonator
    with the given parameters according to Alex Chao's resonator
    model (Eq. 2.82) and definitions of the resonator in HEADTAIL.
    """

    omega = 2 * numpy.pi * frequency
    alpha = omega / (2 * qfactor)
    omegabar = numpy.sqrt(numpy.abs(omega**2 - alpha**2))
    dt = -srange/(beta * clight)
    if qfactor > 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (numpy.cos(omegabar * dt) +
                alpha / omegabar * numpy.sin(omegabar*dt)))
    elif qfactor == 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (1. + alpha * dt))
    elif qfactor < 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (numpy.cosh(omegabar * dt) +
                alpha / omegabar * numpy.sinh(omegabar * dt)))
    return wake


def transverse_resonator_wf(srange, frequency, qfactor, rshunt,
                            yokoya_factor, beta):
    """
    Define the wake function (transverse) of a resonator
    with the given parameters according to Alex Chao's
    resonator model (Eq. 2.82) and definitions of the resonator
    in HEADTAIL.
    """

    omega = 2 * numpy.pi * frequency
    alpha = omega / (2 * qfactor)
    omegabar = numpy.sqrt(numpy.abs(omega**2 - alpha**2))
    dt = -srange/(beta * clight)
    if qfactor > 0.5:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                (qfactor * omegabar) *
                numpy.exp(alpha*dt) * numpy.sin(omegabar*dt))
    elif qfactor == 0.5:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                qfactor *
                numpy.exp(alpha * dt) * dt)
    else:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                (qfactor * omegabar) *
                numpy.exp(alpha*dt) * numpy.sinh(omegabar*dt))
    return wake


def transverse_reswall_wf(srange, yokoya_factor, length, rvac, conduct, beta):
    """
    Define the wake function (transverse) of a resistive wall with the given
    parameters according to Alex Chao's RW model (2.53) and definitions used in
    HEADTAIL
    """

    z0 = 119.9169832 * numpy.pi
    dt = -srange/(beta * clight)
    wake = (yokoya_factor * (numpy.sign(dt) - 1) / 2. *
            beta * length / numpy.pi / rvac**3 *
            numpy.sqrt(-z0 * clight / conduct / numpy.pi / dt))

    wake[srange <= 0] = 0.0

    return wake



"""AT generic plotting function"""
from __future__ import annotations
from itertools import chain, repeat
# noinspection PyPackageRequirements
import matplotlib.pyplot as plt
from typing import Callable
from .synopt import plot_synopt
from ..lattice import Lattice

SLICES = 400

__all__ = ['baseplot']


def baseplot(ring: Lattice, plot_function: Callable, *args, **kwargs):
    """Generic lattice plot

    :py:func:`baseplot` divides the region of interest of ring into small
    elements, calls the specified function to get the plot data and calls
    matplotlib functions to generate the plot.
    By default, it creates a new figure for the plot, but if provided with
    :py:class:`~matplotlib.axes.Axes` objects it can be used as part of a GUI

    Parameters:
        ring:           Lattice description.
        plot_function:  Specific data generating function to be called
          plotting function. ``plot_function`` is called as:

          :code:`title, left, right = plot_function(ring, refpts, *args,
          **kwargs)`

          and should return 2 or 3 output:

          ``title``: plot title or :py:obj:`None`

          ``left``: tuple returning the data for the main (left) axis

            left[0] - y-axis label

            left[1] - xdata: (N,) array (s coordinate)

            left[2] - ydata: iterable of (N,) or (N,M) arrays. Lines from a
            (N, M) array share the same style and label

            left[3]   labels: (optional) iterable of strings as long as ydata

          ``right``: tuple returning the data for the secondary (right) axis
        *args:          All other positional parameters are sent to the

    Keyword Args:
        s_range:            Lattice range of interest, default: unchanged,
          initially set to the full cell.
        axes (tuple[Axes, Optional[Axes]): :py:class:`~matplotlib.axes.Axes`
          for plotting as (primary_axes, secondary_axes).
          Default: create new axes
        slices (int):       Number of slices. Default: 400
        legend (bool):      Show a legend on the plot. Default: :py:obj:`True`
        labels (Refpts):    display the name of selected elements.
          Default: :py:obj:`None`
        block (bool):       If :py:obj:`True`, block until the figure is closed.
          Default: :py:obj:`False`
        dipole (dict):      Dictionary of properties overloading the default
          properties of dipole representation. See :py:func:`.plot_synopt`
          for details
        quadrupole (dict):  Same definition as for dipole
        sextupole (dict):   Same definition as for dipole
        multipole (dict):   Same definition as for dipole
        monitor (dict):     Same definition as for dipole
        **kwargs:           All other keywords are sent to the plotting function

    Returns:
        left_axes (Axes):   Main (left) axes
        right_axes (Axes):  Secondary (right) axes or :py:obj:`None`
        synopt_axes (Axes): Synoptic axes
    """

    def plot1(ax, yaxis_label, x, y, labels=()):
        lines = []
        for y1, prop, label in zip(y, props, chain(labels, repeat(None))):
            ll = ax.plot(x, y1, **prop)
            if label is not None:
                ll[0].set_label(label)
            lines += ll
        ax.set_ylabel(yaxis_label)
        return lines

    def labeled(line):
        return not line.properties()['label'].startswith('_')

    # extract baseplot arguments
    slices = kwargs.pop('slices', SLICES)
    axes = kwargs.pop('axes', None)
    legend = kwargs.pop('legend', True)
    block = kwargs.pop('block', False)
    if 's_range' in kwargs:
        ring.s_range = kwargs.pop('s_range')

    # extract synopt arguments
    synkeys = ['dipole', 'quadrupole', 'sextupole', 'multipole',
               'monitor', 'labels']
    kwkeys = list(kwargs.keys())
    synargs = dict((k, kwargs.pop(k)) for k in kwkeys if k in synkeys)

    # get color cycle
    cycle_props = plt.rcParams['axes.prop_cycle']

    # slice the ring
    rg = ring.slice(slices=slices)

    # get the data for the plot
    pout = plot_function(rg, rg.i_range, *args, **kwargs)
    title = pout[0]
    plots = pout[1:]

    # prepare the axes
    if axes is None:
        # Create new axes
        nplots = len(plots)
        fig = plt.figure()
        axleft = fig.add_subplot(111, xlim=rg.s_range, xlabel='s [m]',
                                 facecolor=[1.0, 1.0, 1.0, 0.0],
                                 title=title)
        axright = axleft.twinx() if (nplots >= 2) else None
        axleft.set_title(ring.name, fontdict={'fontsize': 'medium'},
                         loc='left')
        axsyn = plot_synopt(ring, axes=axleft, **synargs)
    else:
        # Use existing axes
        axleft, axright = axes
        axsyn = None
        nplots = 1 if axright is None else len(plots)

    props = iter(cycle_props())

    # left plot
    lines1 = plot1(axleft, *plots[0])
    # right plot
    lines2 = [] if (nplots < 2) else plot1(axright, *plots[1])
    if legend:
        if nplots < 2:
            axleft.legend(handles=[li for li in lines1 if labeled(li)])
        elif axleft.get_shared_x_axes().joined(axleft, axright):
            axleft.legend(
                handles=[li for li in lines1 + lines2 if labeled(li)])
        else:
            axleft.legend(handles=[li for li in lines1 if labeled(li)])
            axright.legend(handles=[li for li in lines2 if labeled(li)])
    plt.show(block=block)
    return axleft, axright, axsyn



import numpy

'''
This is an example of a python passmethod, here the Drift
passmethod. This is a simplified version that does not check
aperture.

The python passmethods has to be in the python path or in
at.integratorsand contain a function trackFunction()

The element creation is as follows:

drift = elements.Drift('drift', 1.0, PassMethod='pyDriftPass')

To be noted:
The particles are described by a 2 dimensional vector
(6,num_particles)
In case a python passmethod and a c passmethod with the same name
are found, the c passmethod is used
'''


def drift6(r, L):
    p_norm = 1/(1+r[4])
    NormL = L*p_norm
    r[0] += NormL*r[1]
    r[2] += NormL*r[3]
    r[5] += NormL*p_norm*(r[1]*r[1]+r[3]*r[3])/2


def trackFunction(rin, elem=None):
    le = elem.Length
    for rtmp in numpy.atleast_2d(rin.T):
        if hasattr(elem, 'T1'):
            rtmp += elem.T1
        if hasattr(elem, 'R1'):
            rtmp[:] = numpy.dot(elem.R1, rtmp)
        drift6(rtmp, le)
        if hasattr(elem, 'R2'):
            rtmp[:] = numpy.dot(elem.R2, rtmp)
        if hasattr(elem, 'T2'):
            rtmp += elem.T2



import matlab
import numpy as np
import pytest
from numpy.testing import assert_allclose as assert_close

import at


def _ml_refs(refpts, nelems):
    """Convert refpoints to Matlab"""
    uintrefs = at.uint32_refpts(refpts, nelems)
    return matlab.double([float(ref + 1) for ref in uintrefs])


@pytest.mark.parametrize("dp", (-0.01, 0.0, 0.01))
@pytest.mark.parametrize("lattices", ["dba", "hmba"])
def test_avlinopt(engine, request, lattices, dp):
    py_lattice, ml_lattice, _ = request.getfixturevalue(lattices)
    nelems = len(py_lattice)
    refpts = range(nelems)
    mlrefs = _ml_refs(refpts, nelems)

    # Python call
    lindata, avebeta, avemu, avedisp, aves, tune, chrom = at.avlinopt(
        py_lattice, dp, refpts=refpts
    )
    # Matlab call
    ml_data, ml_avebeta, ml_avemu, ml_avedisp, ml_tune, ml_chrom = engine.pyproxy(
        "atavedata", ml_lattice, dp, mlrefs, nargout=6
    )
    # Comparison
    assert_close(avebeta, np.asarray(ml_avebeta), rtol=1e-7)
    assert_close(avemu, np.asarray(ml_avemu), rtol=0, atol=1e-7)
    assert_close(avedisp, np.asarray(ml_avedisp), rtol=0, atol=1e-8)
